"""
Simplified version of harpy from
Jaime Coello Maria de Portugal - Martinez Vazquez
Github: https://github.com/pylhc/harpy
"""

from __future__ import annotations
import numpy
from warnings import warn
from scipy.fft import fft, fftfreq
from at.lattice import AtWarning
from at.lattice import AtError
import multiprocessing
from functools import partial


__all__ = ['get_spectrum_harmonic', 'get_main_harmonic',
           'get_tunes_harmonic']


def _pad_signal(samples, pad_length):
    """ Pad signal and round to the next power of 2"""
    if pad_length is not None:
        length = len(samples)
        pad_length = (1 << (length + int(pad_length)
                            - 1).bit_length()) - length
        samples = numpy.pad(samples, (0, pad_length))
    return samples


def _interpolate_peak(complex_values):
    """This method interpolates the real frequency of the
       signal using the three highest peaks in the FFT.
    """
    k = numpy.argmax(numpy.abs(complex_values))
    rv = complex_values
    n = len(complex_values)
    delta = numpy.tan(numpy.pi / n) / (numpy.pi / n)
    kp = (k + 1) % n
    km = (k - 1) % n
    dk = rv[km] - rv[kp]
    sk = 2 * rv[k] - rv[km] - rv[kp]
    delta *= numpy.real(dk/sk)
    return (k + delta) / n


def _compute_coef(samples, freq):
    """
    Computes the coefficient of the Discrete Time Fourier
    Transform corresponding to the given frequency
    """
    n = len(samples)
    exponents = numpy.exp(-2j*numpy.pi * freq * numpy.arange(n))
    coef = numpy.sum(exponents * samples)/n
    return coef


def _interpolated_fft(samples, num_harmonics, fmin, fmax,
                      maxiter):
    """Compute the interpolated FFT of a signal"""
    rn = numpy.arange(len(samples))
    coefficients = numpy.zeros(num_harmonics, dtype=complex)
    frequencies = numpy.zeros(num_harmonics)

    nfound = 0
    niter = 0
    nmax = num_harmonics * maxiter

    while (nfound < num_harmonics) and (niter < nmax):
        fft_data = fft(samples)
        frequency = _interpolate_peak(fft_data)
        coefficient = _compute_coef(samples, frequency)
        if (frequency >= fmin) and (frequency <= fmax):
            frequencies[nfound] = frequency
            coefficients[nfound] = coefficient
            nfound += 1
        samples = samples - coefficient * numpy.exp(2j*numpy.pi*frequency*rn)
        niter += 1

    if nfound == 0:
        msg = ('No harmonic found within range, '
               'consider extending it or increase maxiter')
        raise AtError(msg)
    elif nfound < num_harmonics:
        msg = ('{0}/{1} harmonics found in '
               'requested range'.format(nfound, num_harmonics))
        warn(AtWarning(msg))

    coefficients, frequencies = zip(*sorted(zip(coefficients, frequencies),
                                    key=lambda tp: numpy.abs(tp[0]),
                                    reverse=True))
    return frequencies, coefficients


def get_spectrum_harmonic(cent: numpy.ndarray, method: str = 'interp_fft',
                          num_harmonics: int = 20,
                          hann: bool = False,
                          fmin: float = 0, fmax: float = 1,
                          maxiter: float = 10,
                          pad_length: float = None,
                          remove_mean: bool = True) -> tuple[numpy.ndarray,
                                                             numpy.ndarray,
                                                             numpy.ndarray]:
    """Frequency analysis of beam motion

    Parameters:
        cent:           Centroid motions of the particle
        method:         ``'interp_fft'`` or ``'interp_fft'``.
                        Default: ``'interp_fft'``
        num_harmonics:  Number of harmonics to search for with interp_fft
        fmin:           Lower bound for spectrum search with interp_fft
        fmax:           Upper bound for spectrum search with interp_fft
        maxiter:        Multiplies ``num_harmonics`` to define the max.
                        number of iteration for the search
        hann:           Turn on Hanning window. Default: :py:obj:`False`.
                        Ignored for interpolated FFT
        pad_length      Zero pad the input signal.
                        Rounded to the higher power of 2
                        Ignored for interpolated FFT
        remove_mean:    Remove the mean of the input signal.
                        Default: :py:obj:`True`.

    Returns:
        frequency (ndarray): (num_harmonics,) array of frequencies
        amplitude (ndarray): (num_harmonics,) array of amplitudes
        phase (ndarray): (num_harmonics,) array of phases
    """
    lc = len(cent)
    # laskar kept for backward compatibility
    if remove_mean:
        cent -= numpy.mean(cent)
    if method == 'interp_fft' or method == 'laskar':
        if hann:
            warn(AtWarning('Windowing not efficient for'
                           'interpolated FFT: ignored'))
        if pad_length is not None:
            warn(AtWarning('Padding not efficient for'
                           'interpolated FFT: ignored'))
        ha_tune, ha_amp = _interpolated_fft(cent, num_harmonics,
                                            fmin, fmax, maxiter)
    elif method == 'fft':
        if hann:
            cent *= numpy.hanning(lc)
        cent = _pad_signal(cent, pad_length)
        fft_data = fft(cent)
        ha_tune, ha_amp = fftfreq(len(fft_data)), fft_data
    else:
        raise ValueError('The method ' + method + ' is undefined')

    ha_phase = numpy.angle(numpy.array(ha_amp))
    ha_amp = numpy.abs(numpy.array(ha_amp)) / lc
    ha_tune = numpy.array(ha_tune)
    return ha_tune, ha_amp, ha_phase


def _get_max_spectrum(freq, amp, phase, fmin, fmax):
    msk = (freq >= fmin) & (freq <= fmax)
    amp = amp[msk]
    freq = freq[msk]
    phase = phase[msk]
    tune = freq[numpy.argmax(amp)]
    phase = phase[numpy.argmax(amp)]
    amp = numpy.amax(amp)
    return tune, amp, phase


def _get_main_single(cents, **kwargs):

    def get_hmain(cents):
        fmin = kwargs.get('fmin', 0)
        fmax = kwargs.get('fmax', 1)
        try:
            out = get_spectrum_harmonic(cents, **kwargs)
            freq, amp, phase = out
            tunes, amps, phases = _get_max_spectrum(freq, amp,
                                                    phase, fmin,
                                                    fmax)
        except AtError:
            msg = ('No harmonic found within range, '
                   'consider extending it or increase maxiter')
            warn(AtWarning(msg))
            tunes = numpy.nan
            amps = numpy.nan
            phases = numpy.nan
        except ValueError:
            msg = ('Invalid input vector provided')
            warn(AtWarning(msg))
            tunes = numpy.nan
            amps = numpy.nan
            phases = numpy.nan
        return tunes, amps, phases

    cents = numpy.atleast_2d(cents)
    results = [get_hmain(c) for c in cents]
    return numpy.transpose(results)


def _get_main_multi(cents, **kwargs):
    cents = numpy.array(cents)
    start_method = kwargs.pop('start_method', None)
    pool_size = kwargs.pop('pool_size', None)
    if cents.ndim > 1:
        npart = cents.shape[0]
    else:
        return _get_main_single(cents, **kwargs)
    if pool_size is None:
        pool_size = min(npart, multiprocessing.cpu_count())
    ctx = multiprocessing.get_context(start_method)
    fun = partial(_get_main_single, **kwargs)
    with ctx.Pool(pool_size) as pool:
        results = pool.map(fun, cents)
    results = numpy.concatenate(results, axis=1)
    return results


def get_main_harmonic(cents: numpy.ndarray, method: str = 'interp_fft',
                      hann: bool = False,
                      fmin: float = 0, fmax: float = 1,
                      num_harmonics: int = 1,
                      maxiter: float = 10,
                      pad_length=None,
                      use_mp: bool = False,
                      pool_size: int = None,
                      start_method: str = None,
                      remove_mean: bool = True) -> tuple[numpy.ndarray,
                                                         numpy.ndarray,
                                                         numpy.ndarray]:
    """Computes tunes, amplitudes and phases from harmonic analysis
    The tune is defined as the harmonic with the maximum amplitude
    within the search range.

    Parameters:
        cents:          Centroid motions of the particle
        method:         ``'interp_fft'`` or ``'fft'``.
                        Default: ``'interp_fft'``
        fmin:           Lower bound for tune search
        fmax:           Upper bound for tune search
        num_harmonics:  Number of harmonics to search for.
                        Default=1.
        maxiter:        Maximum number of iterations for the search
        hann:           Turn on Hanning window. Default: :py:obj:`False`.
                        Ignored for interpolated FFT
        pad_length:     Zero pad the input signal.
                        Rounded to the higher power of 2
                        Ignored for interpolated FFT
        use_mp (bool): Flag to activate multiprocessing (default: False)
        pool_size:     number of processes used when
          *use_mp* is :py:obj:`True`. If None, ``min(npart,nproc)``
          is used. It can be globally set using the variable
          *at.lattice.DConstant.patpass_poolsize*
        start_method:  Python multiprocessing start method. Default None
                       uses the OS default method.
        remove_mean:    Remove the mean of the input signal.
                        Default: :py:obj:`True`.

    Returns:
        tunes (ndarray):    numpy array of length len(cents), max of the
        spectrum within [fmin fmax]
        amplitude (ndarray): (len(cents), ) array of amplitudes
                             corresponding to the tune
        phase (ndarray): (len(cents), ) array of phases
                         corresponding to the tune


    .. note::

       * The tune is defined as the harmonic with the maximum amplitude within
         the search range ``(fmin, fmax)``.
       * In case a ``Nan`` is present in the input vector or the tune cannot
         be found within the range, the function returns ``NaN``.
       * For the method ``'interp_fft'``, harmonics are calculated iteratively
         starting from the maximum peak of the raw FFT. ``num_harmonics=1``
         is the default, only the first harmonic is calculated.
         However, it is possible that the maximum of the interpolated
         FFT does not correspond to the maximum of the raw FFT, in which case
         ``num_harmonics`` has to be increased to get the correct peak.
    """
    if use_mp:
        tunes, amps, phases = _get_main_multi(cents,
                                              num_harmonics=num_harmonics,
                                              method=method, hann=hann,
                                              pad_length=pad_length,
                                              fmin=fmin, fmax=fmax,
                                              maxiter=maxiter,
                                              pool_size=pool_size,
                                              start_method=start_method,
                                              remove_mean=remove_mean)
    else:
        tunes, amps, phases = _get_main_single(cents,
                                               num_harmonics=num_harmonics,
                                               method=method, hann=hann,
                                               pad_length=pad_length,
                                               fmin=fmin, fmax=fmax,
                                               maxiter=maxiter,
                                               remove_mean=remove_mean)
    return tunes, amps, phases


def get_tunes_harmonic(cents: numpy.ndarray, method: str = 'interp_fft',
                       hann: bool = False,
                       fmin: float = 0, fmax: float = 1,
                       num_harmonics: int = 1,
                       maxiter: float = 10,
                       pad_length=None,
                       use_mp: bool = False,
                       pool_size: int = None,
                       start_method: str = None,
                       remove_mean: bool = True,
                       **kwargs) -> numpy.ndarray:
    """Computes tunes from harmonic analysis.

    Parameters:
        cents:          Centroid motions of the particle
        method:         ``'interp_fft'`` or ``'fft'``.
                        Default: ``'interp_fft'``
        fmin:           Lower bound for tune search
        fmax:           Upper bound for tune search
        num_harmonics:  Number of harmonics to search for.
                        Default=1.
        maxiter:        Maximum number of iterations for the search
        hann:           Turn on Hanning window. Default: :py:obj:`False`.
                        Ignored for interpolated FFT.
        pad_length:     Zero pad the input signal.
                        Rounded to the higher power of 2
                        Ignored for interpolated FFT
        use_mp (bool): Flag to activate multiprocessing (default: False)
        pool_size:     number of processes used when
          *use_mp* is :py:obj:`True`. If None, ``min(npart,nproc)``
          is used. It can be globally set using the variable
          *at.lattice.DConstant.patpass_poolsize*
        start_method:  Python multiprocessing start method. Default None
                       uses the OS default method.
        remove_mean:    Remove the mean of the input signal.
                        Default: :py:obj:`True`.

    Returns:
        tunes (ndarray):    numpy array of length len(cents), max of the
        spectrum within [fmin fmax]

    .. note::

       * The tune is defined as the harmonic with the maximum amplitude within
         the search range ``(fmin, fmax)``.
       * In case a ``Nan`` is present in the input vector or the tune cannot
         be found within the range, the function returns ``NaN``.
       * For the method ``'interp_fft'``, harmonics are calculated iteratively
         starting from the maximum peak of the raw FFT. ``num_harmonics=1``
         is the default, only the first harmonic is calculated.
         However, it is possible that the maximum of the interpolated
         FFT does not correspond to the maximum of the raw FFT, in which case
         ``num_harmonics`` has to be increased to get the correct peak.
    """
    tunes, _, _ = get_main_harmonic(cents, method=method, hann=hann, fmin=fmin,
                                    fmax=fmax, pad_length=pad_length,
                                    use_mp=use_mp, pool_size=pool_size,
                                    start_method=start_method, maxiter=maxiter,
                                    remove_mean=remove_mean,
                                    num_harmonics=num_harmonics)
    return tunes



import numpy
import pytest

from at.load.tracy import (
    tokenise_expression,
    parse_float,
    expand_tracy,
    parse_lines,
    tracy_element_from_string,
)
from at.lattice.elements import (
    Dipole,
    Drift,
    Marker,
    Multipole,
    Quadrupole,
    RFCavity,
    Sextupole,
)


@pytest.fixture
def defaults():
    return {"energy": 3e9, "harmonic_number": 936}


@pytest.mark.parametrize(
    "exp,target",
    [
        ["3", ["3"]],
        ["-3", ["-", "3"]],
        ["-4.85d-3", ["-", "4.85d-3"]],
        ["3+4", ["3", "+", "4"]],
        ["3 + 4", ["3", "+", "4"]],
        ["3+4e3", ["3", "+", "4e3"]],
        ["1e3+4e-3", ["1e3", "+", "4e-3"]],
        ["1 - (3 / 2)", ["1", "-", "(", "3", "/", "2", ")"]],
        ["(1 - 1e-1) + 3", ["(", "1", "-", "1e-1", ")", "+", "3"]],
        ["0.485-0.002", ["0.485", "-", "0.002"]],
        [" (3.482-0.02)-1.85", ["(", "3.482", "-", "0.02", ")", "-", "1.85"]],
        ["tbx-tbev-tbes", ["tbx", "-", "tbev", "-", "tbes"]],
    ],
)
def test_tokenise_expression(exp, target):
    assert tokenise_expression(exp) == target


@pytest.mark.parametrize(
    "exp,target",
    [
        ["3.4", 3.4],
        ["-3", -3],
        ["3+4", 7],
        ["3.8*1.1", 4.18],
        ["3+4e3", 4003],
        ["1e3+4e-3", 1000.004],
        ["1 + 12 / (a + 1)", 5],
        ["0.485-0.002", 0.483],
        ["4.85d3", 4850],
        ["-4.85d-3", -0.00485],
    ],
)
def test_parse_float(exp, target):
    assert parse_float(exp, {"a": 2}) == target


def test_parse_lines_removes_comments():
    assert parse_lines("a{}b") == ["ab"]


def test_parse_lines_removes_whitespace():
    assert parse_lines("a\n{a\nb}b") == ["ab"]


def test_parse_line_splits_on_semicolons():
    assert parse_lines("a;b") == ["a", "b"]


def test_tracy_element_from_string_handles_drift(defaults):
    drift = "drift,l=0.0450000"
    assert tracy_element_from_string("d1", drift, defaults).equals(
        Drift("d1", 0.0450000)
    )


def test_tracy_element_from_string_handles_marker(defaults):
    marker = "marker"
    expected = Marker("m1")
    assert tracy_element_from_string("m1", marker, defaults).equals(expected)


def test_tracy_element_from_string_handles_quadrupole(defaults):
    quad = "quadrupole,l=0.15,k=10.24,n=nquad,method=4"
    defaults["nquad"] = 10
    expected = Quadrupole(
        "q1",
        0.15,
        k=10.24,
        NumIntSteps=10,
        method="4",
        PassMethod="StrMPoleSymplectic4Pass",
    )
    assert tracy_element_from_string("q1", quad, defaults).equals(expected)


def test_tracy_element_from_string_handles_sextupole(defaults):
    sext = "sextupole,l=0.14,k=174.4,n=nsext,method=4"
    defaults["nsext"] = 2
    expected = Sextupole("s1", 0.14, h=174.4, NumIntSteps=2, method="4")
    assert tracy_element_from_string("s1", sext, defaults).equals(expected)


def test_tracy_element_from_string_handles_hom(defaults):
    oct = "multipole,l=0.0,hom=(4,1.0,0.3)"
    expected = Multipole("m1", 0.0, poly_a=[0, 0, 0, 0.3], poly_b=[0, 0, 0, 1])
    assert tracy_element_from_string("m1", oct, defaults).equals(expected)


def test_tracy_element_from_string_handles_cavity(defaults):
    cavity = "cavity,l=0.0,frequency=499.654e6,voltage=2.2e6,phi=0.0"
    expected = RFCavity("c1", 0.0, 2.2e6, 4.99654e8, 936, 3e9, Phi=0.0)
    constructed = tracy_element_from_string("c1", cavity, defaults)
    assert constructed.equals(expected)


def test_tracy_element_from_string_handles_bending(defaults):
    bend = "bending,l= 0.20000000,t=0.32969999,t1=0.00000000," \
           "t2=0.32969999,k=-0.124,gap=0.06,n=nbend,method=4"
    defaults["nbend"] = 2
    expected = Dipole(
        "b1",
        0.2,
        BendingAngle=0.00575435036929238,
        EntranceAngle=0,
        ExitAngle=0.00575435036929238,
        k=-0.12411107,
        FullGap=0.03,
        FringeInt1=1,
        FringeInt2=1,
        NumIntSteps=2,
        PolynomA=numpy.array([0, 0, 0, 0], dtype=numpy.float64),
        PolynomB=numpy.array([0, -0.124, 0, 0], dtype=numpy.float64),
        method="4",
        PassMethod="BndMPoleSymplectic4Pass",
    )
    assert tracy_element_from_string("b1", bend, defaults).equals(expected)


def test_tracy_element_from_string_handles_variable(defaults):
    drift = "drift,l=a"
    defaults["a"] = 1
    correct_drift = Drift("d1", 1)
    constructed_drift = tracy_element_from_string("d1", drift, defaults)
    assert correct_drift.equals(constructed_drift)


def test_expand_tracy():
    contents = "define lattice;energy=1;dmult:drift,l=1;cell:dmult;end;"
    elements, energy = expand_tracy(contents, "cell", 936)
    assert len(elements) == 1
    assert energy == 1e9
    assert elements[0].equals(Drift("dmult", 1.0))



"""
Classes for matching variables and constraints
"""
from __future__ import annotations
from itertools import chain
import numpy as np
from collections.abc import Sequence, Callable
from typing import Optional, Union
from scipy.optimize import least_squares
from itertools import repeat
from at.lattice import Lattice, Refpts, bool_refpts
from at.physics import get_optics, ohmi_envelope, find_orbit


class Variable(object):
    """A :py:class:`Variable` is a scalar value acting on a lattice through the
    user-defined functions setfun and getfun

    Parameters:
        setfun:     User-defined function for setting the Variable. Called as:

          :code:`setfun(ring, value, *args, **kwargs)`

          where :code:`value` is the scalar value to apply

          The positional and keyword parameters come from the
          :py:class:`Variable` initialisation
        getfun:     User-defined function for retrieving the actual value of
          the variable: Called as:

          :code:`value = getfun(ring, *args, **kwargs)`

          The positional and keyword parameters come from the
          :py:class:`Variable` initialisation
        name:       Name of the Variable; Default: ``''``
        bounds:     Lower and upper bounds of the variable value
        fun_args:       Positional arguments transmitted to ``setfun`` and
          ``getfun`` functions

    Keyword Args:
        **fun_kwargs:   Keyword arguments transmitted to ``setfun``and
          ``getfun`` functions
    """
    def __init__(self, setfun: Callable, getfun: Callable,
                 name: str = '',
                 bounds: tuple[float, float] = (-np.inf, np.inf),
                 fun_args: tuple = (), **fun_kwargs):
        self.setfun = setfun
        self.getfun = getfun
        self.name = name
        self.bounds = bounds
        self.args = fun_args
        self.kwargs = fun_kwargs
        super(Variable, self).__init__()

    def set(self, ring: Lattice, value):
        """Set the variable value"""
        self.setfun(ring, value, *self.args, **self.kwargs)

    def get(self, ring: Lattice):
        """Get the actual variable value"""
        return self.getfun(ring, *self.args, **self.kwargs)

    @staticmethod
    def header():
        return '\n{:>12s}{:>13s}{:>16s}{:>16s}\n'.format(
            'Name', 'Initial', 'Final ', 'Variation')

    def status(self, ring: Lattice, vini=np.nan):
        vnow = self.get(ring)
        return '{:>12s}{: 16e}{: 16e}{: 16e}'.format(
            self.name, vini, vnow, (vnow - vini))


class ElementVariable(Variable):
    """An :py:class:`ElementVariable` is:

    * a scalar attribute or
    * an element of an array attribute

    of one or several :py:class:`.Element` (s) of a lattice.

    Parameters:
        refpts:     Location of variable :py:class:`.Element` (s)
        attname:    Attribute name
        index:      Index in the attribute array. Use :py:obj:`None` for
          scalar attributes
        name:       Name of the Variable; Default: ``''``
        bounds:     Lower and upper bounds of the variable value
    """

    def __init__(self, refpts: Refpts, attname: str,
                 index: Optional[int] = None,
                 name: str = '',
                 bounds: tuple[float, float] = (-np.inf, np.inf)):
        setf, getf = self._access(index)

        def setfun(ring, value):
            for elem in ring.select(refpts):
                setf(elem, attname, value)

        def getfun(ring):
            values = np.array([getf(elem, attname) for elem in
                               ring.select(refpts)])
            return np.average(values)

        super(ElementVariable, self).__init__(setfun, getfun, name=name,
                                              bounds=bounds)
        self.refpts = refpts

    @staticmethod
    def _access(index):
        """Access to element attributes"""
        if index is None:
            setf = setattr
            getf = getattr
        else:
            def setf(elem, attrname, value):
                getattr(elem, attrname)[index] = value

            def getf(elem, attrname):
                return getattr(elem, attrname)[index]
        return setf, getf


class Constraints(object):
    """Container for generic constraints:

    * a constraint is defined by a user-defined evaluation function,
    * constraints are added to the container with the :py:meth:`add` method.

    Parameters:
        *args:      Positional arguments sent to the evaluation functions
          of all the embedded constraints

    Keyword Args:
        **kwargs:   Keyword arguments sent to the evaluation functions
          of all the embedded constraints

    Examples:
        Define an evaluation function for the ring circumference:

        >>> def circ_fun(ring):
        ...     return ring.get_s_pos(len(ring) + 1)

        Define an evaluation function for the momentum compaction factor:

        >>> def mcf_fun(ring):
        ...     return ring.get_mcf()

        Construct the container:

        >>> cnstrs = Constraints()

        Add the two constraints:

        >>> cnstrs.add(circ_fun, 850.0)
        >>> cnstrs.add(mcf_fun, 1.0e-4, weight=0.1)
    """
    def __init__(self, *args, **kwargs):
        self.name = []
        self.fun = []
        self.target = []
        self.weight = []
        self.lbound = []
        self.ubound = []
        self.rad = kwargs.pop('rad', False)
        self.args = args
        self.kwargs = kwargs

    def add(self, fun: Callable, target, name: Optional[str] = None,
            weight=1.0, bounds=(0.0, 0.0)):
        """Add a target to the :py:class:`Constraints` container

        .. highlight:: python

        Parameters:
            fun:          Evaluation function. Called as:

              :code:`value = fun(ring, *args, **kwargs)`

              ``value`` is the constrained parameter value,

              ``value`` may be a scalar or an array.

              The positional and keyword parameters come from
              the :py:class:`Constraints` initialisation
            target:       Desired value.
            name:         Name of the constraint. If :py:obj:`None`, a ``name``
              is generated from the name of the evaluation function
            weight:       Weight factor: the residual is
              :code:`(value-target)/weight`
            bounds:       Lower and upper bounds. The parameter is constrained
              in the interval [target-low_bound target+up_bound]

        The "target", "weight" and "bounds" input must be broadcastable to the
        shape of "value".
        """
        if name is None:                # Generate the constraint name
            name = fun.__name__
        self.name.append(name)
        self.fun.append(fun)
        target, lbound, ubound = np.atleast_1d(target, *bounds)
        self.target.append(target)
        self.weight.append(weight)
        self.lbound.append(lbound)
        self.ubound.append(ubound)

    def values(self, ring: Lattice):
        """Return the list of actual parameter values"""
        return [fun(ring, *self.args, **self.kwargs) for fun in self.fun]

    def evaluate(self, ring: Lattice):
        """Return a flattened array of weighted residuals"""

        def res(value, target, weight, lbound, ubound):
            diff = value - target   # broadcast here => at least 1 dimension
            lb = diff + lbound
            ub = diff - ubound
            lb[lb >= 0] = 0         # needs at least 1 dimension
            ub[ub <= 0] = 0
            return np.maximum(abs(lb), abs(ub)) / weight

        return np.concatenate([res(v, t, w, lb, ib) for v, t, w, lb, ib
                               in zip(self.values(ring), self.target,
                                      self.weight, self.lbound,
                                      self.ubound)], axis=None)

    @staticmethod
    def header():
        """Header for the display of constraint values and residuals"""
        return '\n{:>12s}{:>13s}{:>16s}{:>16s}{:>16s}\n'.format(
            'Name', 'Initial', 'Final ', 'Target', 'Residual')

    def status(self, ring: Lattice, initial=None):
        """Return a string giving the actual state of constraints"""
        if initial is None:
            initial = repeat(np.nan)
        strs = []
        for name, ini, now, target in zip(self.name, initial,
                                          self.values(ring), self.target):
            now, target = np.broadcast_arrays(now, target)
            for vi, vn, vt in zip(np.broadcast_to(ini, now.shape).flat,
                                  now.flat,
                                  np.broadcast_to(target, now.shape).flat):
                strs.append('{:>12s}{: 16e}{: 16e}{: 16e}{: 16e}'.format(
                    name, vi, vn, vt, vn - vt))
        return '\n'.join(strs)


class ElementConstraints(Constraints):
    """Base class for position-related constraints: handle the refpoints
    of each target"""
    def __init__(self, ring: Lattice, *args, **kwargs):
        self.nelems = len(ring)
        self.refs = []
        self.refpts = ring.bool_refpts([])
        super(ElementConstraints, self).__init__(*args, **kwargs)

    @staticmethod
    def _arrayaccess(index):
        """Access to array elements"""
        if index is None:
            def getv(x):
                return x
        else:
            def getv(x):
                return x[index]
        return getv

    @staticmethod
    def _recordaccess(index):
        """Access to optics parameters"""
        if index is None:
            getf = getattr
        else:
            if isinstance(index, tuple):
                idx = (Ellipsis,)+index
            else:
                idx = (Ellipsis, index)

            def getf(lindata, attrname):
                return getattr(lindata, attrname)[idx]
        return getf

    def add(self, fun: Callable, target,
            refpts: Optional[Refpts] = None, **kwargs):
        """Add a target to the :py:class:`ElementConstraints` container

        Parameters:
            fun:          Evaluation function. Called as:

              :code:`value = fun(ring, *args, **kwargs)`

              ``value`` is the constrained parameter value

              ``value`` may be a scalar or an array.

              the positional and keyword parameters come from
              the :py:class:`ElementConstraints` initialisation
            target:       Desired value.
            refpts:       Location of the constraint

        Keyword Args:
            name:         Name of the constraint. If :py:obj:`None`, a ``name``
              is generated from the name of the evaluation function
            weight:       Weight factor: the residual is
              :code:`(value-target)/weight`
            bounds:       Lower and upper bounds. The parameter is constrained
              in the interval [target-low_bound target+up_bound]

        The "target", "weight" and "bounds" input must be broadcastable to the
        shape of "value".
        """
        ref = bool_refpts(refpts, self.nelems)
        # Store the new refpoint
        self.refs.append(ref)
        # Update the union of all refpoints
        self.refpts = np.stack((self.refpts, ref), axis=0).any(axis=0)
        super(ElementConstraints, self).add(fun, target, **kwargs)

    def values(self, ring: Lattice):
        # Single optics computation
        vloc, vglob = self.compute(ring, *self.args, **self.kwargs)
        # Evaluate all constraints
        return [fun(loc, *vglob) for fun, loc in zip(self.fun, vloc)]

    def compute(self, ring: Lattice, *args, **kwargs):
        """Dummy computation. Compute must return:
        - an iterator over local data for each target
        - a tuple of global data
        """
        return repeat(None), ()


class LinoptConstraints(ElementConstraints):
    # noinspection PyUnresolvedReferences
    """Container for linear optics constraints:

    * a constraint can be set on any result of at.get_optics
    * constraints are added to the container with the LinoptConstraints.add
      method.

    :py:func:`.get_optics` is called once before the evaluation of all
    constraints

    Parameters:
        ring:       Lattice description

    Keyword Args:
        dp (float):   Momentum deviation.
        dct (float):  Path lengthening. If specified, ``dp`` is
          ignored and the off-momentum is deduced from the path lengthening.
        orbit (Optional[Orbit]): Avoids looking for the closed orbit if is
          already known ((6,) array)
        twiss_in:   Initial twiss parameters for transfer line optics.
          See :py:func:`.linopt6`
        method (Callable):  Method used for the analysis of the
          transfer matrix. Can be :py:obj:`~.linear.linopt2`,
          :py:obj:`~.linear.linopt4`, :py:obj:`~.linear.linopt6`

          * :py:obj:`~.linear.linopt2`: No longitudinal motion,
                                        no H/V coupling,
          * :py:obj:`~.linear.linopt4`: No longitudinal motion, Sagan/Rubin
            4D-analysis of coupled motion,
          * :py:obj:`~.linear.linopt6` (default): With or without longitudinal
            motion, normal mode analysis

    Example:

        >>> cnstrs = LinoptConstraints(ring, dp=0.01, coupled=False)

        Add a beta x (beta[0]) constraint at location ref_inj:

        >>> cnstrs.add('beta', 18.0, refpts=ref_inj,
                       name='beta_x_inj', index=0)

        Add an horizontal tune (tunes[0]) constraint:

        >>> cnstrs.add('tunes', 0.44, index=0, weight=0.01)

        Add a chromaticity constraint (both planes):

        >>> cnstrs.add('chroms', [0.0 0.0])

        Define a constraint of phase advances between 2 points:

        >>> def mu_diff(lindata, tune, chrom):
        ...     delta_mu = (lindata[1].mu - lindata[0].mu)/(2*np.pi)
        ...     return delta_mu % 1.0

        Add a H phase advance constraint, giving the desired locations:

        >>> cnstrs.add(mu_diff, 0.5, refpts=[sf0 sf1], index=0)
        """
    def __init__(self, ring: Lattice, **kwargs):
        self.get_chrom = False
        super(LinoptConstraints, self).__init__(ring, **kwargs)

    def add(self, param, target, refpts: Optional[Refpts] = None,
            index: Optional[Union[int, slice]] = None,
            name: Optional[str] = None, **kwargs):
        """Add a target to the LinoptConstraints container

        Parameters:
            param:         2 possibilities:

              * parameter name: see :py:func:`.linopt6` for the
                name of available parameters. In addition to local
                optical parameters, ``'tunes'`` and ``'chroms'``
                are allowed.
              * user-supplied parameter evaluation function:

                :code:`value = param(lindata, tune, chrom)`

                ``lindata`` contains the optics parameters at all the
                specified refpoints
                ``value`` is the constrained parameter value
                (scalar or array).
            target:       desired value.
            refpts:       location of the constraint. Several locations may be
              given to apply the same constraint at several points.
            index:        index in the parameter array. If :py:obj:`None`,
              the full array is used.
            name:         name of the constraint. If :py:obj:`None`, name is
              generated from ``param`` and ``index``.

        Keyword Args:
            weight:       Weight factor: the residual is
              :code:`(value-target)/weight`
            bounds:       lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]
            UseInteger:   Match integer part of mu, much slower as the optics
                          calculation is done for all refpts

        The target, weight and bounds values must be broadcastable to the shape
        of value.
        """
        getf = self._recordaccess(index)
        getv = self._arrayaccess(index)
        use_integer = kwargs.pop('UseInteger', False)
        norm_mu = {'mu': 1, 'mun': 2*np.pi}

        if name is None:                # Generate the constraint name
            name = param.__name__ if callable(param) else param
            if index is not None:
                name = '{}_{}'.format(name, index)

        if callable(param):
            def fun(refdata, tune, chrom):
                return getv(param(refdata, tune, chrom))
            self.get_chrom = True       # fun may use dispersion or chroma
        elif param == 'tunes':
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                return getv(tune)
            refpts = []
        elif param == 'chroms':
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                return getv(chrom)
            refpts = []
            self.get_chrom = True       # slower but necessary
        elif param == 'mu' or param == 'mun':
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                if use_integer:
                    return getf(refdata, 'mu') / norm_mu[param]
                else:
                    return (getf(refdata, 'mu') % (2*np.pi)) / norm_mu[param]
            if use_integer:
                self.refpts[:] = True  # necessary not to miss 2*pi jumps
            else:
                target = target % (2 * np.pi / norm_mu[param])
        else:
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                return getf(refdata, param)

        super(LinoptConstraints, self).add(fun, target, refpts, name=name,
                                           **kwargs)

    def compute(self, ring: Lattice, *args, **kwargs):
        """Optics computation before evaluation of all constraints"""
        ld0, bd, ld = get_optics(ring, refpts=self.refpts,
                                 get_chrom=self.get_chrom, **kwargs)
        return (ld[ref[self.refpts]] for ref in self.refs), \
               (bd.tune[:2], bd.chromaticity)


class OrbitConstraints(ElementConstraints):
    # noinspection PyUnresolvedReferences
    """Container for orbit constraints:
    The closed orbit can be handled with :py:class`LinoptConstraints`, but for
    problems which do not involve parameters other than orbit, like steering or
    orbit bumps, :py:class:`OrbitConstraints` is much faster.

    :py:func:`.find_orbit` is called once before the evaluation of all
    constraints

    Parameters:
        ring:       Lattice description

    Keyword Args:
        dp (float):   Momentum deviation.
        dct (float):  Path lengthening. If specified, ``dp`` is
          ignored and the off-momentum is deduced from the path lengthening.
        orbit (Optional[Orbit]): Avoids looking for the closed orbit if is
          already known ((6,) array)

    Example:
        >>> cnstrs = OrbitConstraints(ring, dp=0.01)

        Add a bump (x=-0.004, x'=0) constraint at location ref_inj

        >>> cnstrs.add([-0.004, 0.0], refpts=ref_inj, index=slice(2))
    """
    def __init__(self, ring: Lattice, *args, **kwargs):
        if ring.radiation:
            kwargs.pop('dp', 0.0)
            kwargs.pop('dct', 0.0)
        super(OrbitConstraints, self).__init__(ring, *args, **kwargs)

    def add(self, target, refpts: Optional[Refpts] = None,
            index: Optional[Union[int, slice]] = None,
            name: Optional[str] = None, **kwargs):
        """Add a target to the OrbitConstraints container

        Parameters:
            target:       desired value.
            refpts:       location of the constraint. Several locations may be
              given to apply the same constraint at several points.
            index:        index in the orbit vector. If :py:obj:`None`, the
              full orbit is used.
            name:         name of the constraint. Default: ``'orbit'``

        Keyword Args:
            weight:       Weight factor: the residual is
              :code:`(value-target)/weight`
            bounds:       lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]

        The target, weight and bounds values must be broadcastable to the shape
        of value.
        """

        if name is None:                # Generate the constraint name
            name = 'orbit' if index is None else 'orbit_{}'.format(index)

        fun = self._arrayaccess(index)
        super(OrbitConstraints, self).add(fun, target, refpts, name=name,
                                          **kwargs)

    def compute(self, ring: Lattice, *args, **kwargs):
        """Orbit computation before evaluation of all constraints"""
        orbit0, orbit = find_orbit(ring, refpts=self.refpts, **kwargs)
        return (orbit[ref[self.refpts]].T for ref in self.refs), ()


class EnvelopeConstraints(ElementConstraints):
    """Container for envelope constraints:

    * a constraint can be set on any result of :py:func:`.ohmi_envelope`,
    * constraints are added to the container with the :py:meth:`add` method.

    :py:func:`.ohmi_envelope` is called once before the evaluation of all
    constraints.

    Parameters:
        ring:       Lattice description
    """
    def __init__(self, ring: Lattice):
        super(EnvelopeConstraints, self).__init__(ring, rad=True)

    def add(self, param, target, refpts: Optional[Refpts] = None,
            index: Optional[Union[int, slice]] = None,
            name: Optional[str] = None, **kwargs):
        """Add a target to the :py:class:`EnvelopeConstraints` container

        Parameters:
            param:        2 possibilities:

              - parameter name: see :py:func:`.ohmi_envelope` for the
                name of available parameters. In addition to local parameters,
                ``'tunes'``, ``'damping_rates'``, ``'mode_matrices'`` and
                ``'mode_emittance'`` are allowed.
              - user-supplied parameter evaluation function:

                :code:`value = param(emit_data, beam_data)`

                ``emit_data`` contains the emittance data at all the
                specified refpoints
                ``value`` is the constrained parameter value
                (scalar or array).
            target:       desired value.
            refpts:       location of the constraint. Several locations may be
              given to apply the same constraint at several points.
            index:        index in the parameter array. If :py:obj:`None`,
              the full array is used.
            name:         name of the constraint. If :py:obj:`None`, name is
              generated from ``param`` and ``index``.

        Keyword Args:
            weight:       Weight factor: the residual is
              :code:`(value-target)/weight`
            bounds:       lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]

        The target, weight and bounds values must be broadcastable to the shape
        of value.
        """
        def beam_constraint(prm):
            # noinspection PyUnusedLocal
            def beamfunc(refdata, beam_data):
                return getv(beam_data[prm])
            return beamfunc

        getf = self._recordaccess(index)
        getv = self._arrayaccess(index)

        if name is None:                # Generate the constraint name
            name = param.__name__ if callable(param) else param
            if index is not None:
                name = '{}_{}'.format(name, index)

        if callable(param):
            def fun(refdata, beam_data):
                return getv(param(refdata, beam_data))
        elif param in ['tunes', 'damping_rates', 'mode_matrices',
                       'mode_emittances']:
            fun = beam_constraint(param)
            refpts = []
        else:
            # noinspection PyUnusedLocal
            def fun(refdata, beam_data):
                return getf(refdata, param)

        super(EnvelopeConstraints, self).add(fun, target, refpts, name=name,
                                             **kwargs)

    def compute(self, ring: Lattice, *args, **kwargs):
        """Optics computation before evaluation of all constraints"""
        em0, beamdata, em = ohmi_envelope(ring, refpts=self.refpts, **kwargs)
        return (em[ref[self.refpts]] for ref in self.refs), (beamdata,)


def match(ring: Lattice, variables: Sequence[Variable],
          constraints: Sequence[Constraints], verbose: int = 2,
          max_nfev: int = 1000,
          diff_step: float = 1.0e-10,
          method=None, copy: bool = True):
    """Perform matching of constraints by varying variables

    Parameters:
        ring:               Lattice description
        variables:          sequence of Variable objects
        constraints:        sequence of Constraints objects
        verbose:            Print additional information
        max_nfev:           Maximum number of evaluations
        diff_step:          Convergence threshold
        method:
        copy:
    """
    def fun(vals):
        for value, variable in zip(vals, variables):
            variable.set(ring1, value)

        c = [cons.evaluate(ring1) for cons in constraints]
        return np.concatenate(c, axis=None)

    if copy:
        # Make a shallow copy of ring
        ring1 = ring.copy()
        varpts = ring.bool_refpts([])
        # build the list of variable elements
        for var in variables:
            if isinstance(var, ElementVariable):
                varpts |= ring.bool_refpts(var.refpts)
        # make a deep copy of all the variable elements
        for ref in ring.uint32_refpts(varpts):
            ring1[ref] = ring1[ref].deepcopy()
    else:
        ring1 = ring

    aaa = [(var.get(ring1), var.bounds) for var in variables]
    vini, bounds = zip(*aaa)
    bounds = np.array(bounds).T

    cini = [cst.values(ring1) for cst in constraints]
    ntargets = sum(np.size(a) for a in chain.from_iterable(cini))

    if method is None:
        if np.all(abs(bounds) == np.inf) and ntargets >= len(variables):
            method = 'lm'
        else:
            method = 'trf'
    if verbose >= 1:
        print('\n{} constraints, {} variables, using method {}\n'.
              format(ntargets, len(variables), method))

    least_squares(fun, vini, bounds=bounds, verbose=verbose, max_nfev=max_nfev,
                  method=method, diff_step=diff_step)

    if verbose >= 1:
        print(Constraints.header())
        for cst, ini in zip(constraints, cini):
            print(cst.status(ring1, initial=ini))

        print(Variable.header())
        for var, vini in zip(variables, vini):
            print(var.status(ring1, vini=vini))

    return ring1



import at
from at.collective import Wake, LongResonatorElement
from mpi4py import MPI
import sys
import pickle as pkl


class Logger(object):
    def __init__(self, fName):

        self.terminal = sys.stdout
        self.log = open(fName, "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass


def launch():

    sys.stdout = Logger('./logfiler.txt')
    sys.stderr = Logger('./logfiler.txt')

    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()

    print(size, rank)

    # Set up the ring
    ring = at.load_m('../../../machine_data/esrf.m')

    nturns = 5000
    Nbunches = 992

    # Number of particles PER CORE must be a multiple of Nbunches
    Npart = int(Nbunches)

    current = 200e-3

    bunchSpacing = ring.circumference / Nbunches
    ring.beam_current = current
    ring.set_rf_voltage(6e6)

    ring.disable_6d(cavity_pass='RFCavityPass')
    ring.set_fillpattern(Nbunches)

    ring.set_rf_frequency()

    qx, qy, qs = ring.get_tune()
    fs0 = qs * ring.revolution_frequency

    ld0, ld = ring.find_orbit()
    fring, _ = at.fast_ring(ring)

    # Define the resonator parameters and current
    wturns = 40
    srange = Wake.build_srange(0., 0.41, 1.0e-3, 1.0e-1,
                               bunchSpacing, ring.circumference*wturns)

    fres = ring.get_rf_frequency() - ring.revolution_frequency - fs0
    qfactor = 1e4
    rshunt = 2e6

    welem = LongResonatorElement('wake', ring, srange, fres, qfactor,
                                 rshunt, Nturns=wturns, Nslice=1)
    fring.append(welem)

    bmon = at.BeamMoments('mon')
    fring.append(bmon)

    # Particle generation and tracking
    sigm = at.sigma_matrix(ring.enable_6d(copy=True))
    part = at.beam(Npart, sigm, orbit=ld0)
    part[4, :] = 0
    part[5, :] = 0

    fring.track(part, nturns=nturns, refpts=None, in_place=True)

    dp_all = bmon.means[4, :, :]
    z_all = bmon.means[5, :, :]
    if rank == 0:
        outdict = {'dp': dp_all.T, 'z': z_all.T,
                   'ring': ring, 'fres': fres, 'qfactor': qfactor,
                   'rshunt': rshunt, 'M': Nbunches,
                   'current': current}
        pkl.dump(outdict, open('./LCBI_output.pkl', 'wb'))


if __name__ == '__main__':
    launch()



from ..lattice import Lattice, get_rf_frequency, check_6d, get_s_pos
from ..lattice import DConstant
from ..constants import clight
from ..tracking import internal_lpass
from .orbit import find_orbit4
import numpy
from typing import Optional

__all__ = ['get_mcf', 'get_slip_factor', 'get_revolution_frequency',
           'set_rf_frequency']


@check_6d(False)
def get_mcf(ring: Lattice, dp: Optional[float] = 0.0,
            keep_lattice: bool = False,
            fit_order: Optional[int] = 1,
            n_step: Optional[int] = 2,
            **kwargs) -> float:
    r"""Compute the momentum compaction factor :math:`\alpha`

    Parameters:
        ring:           Lattice description (:pycode:`ring.is_6d` must be
          :py:obj:`False`)
        dp:             Momentum deviation. Defaults to :py:obj:`None`
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: :py:obj:`False`
        fit_order:      Maximum momentum compaction factor order to be fitted.
          Default to 1, corresponding to the first-order momentum compaction factor.
        n_step:         Number of different calculated momentum deviations to be fitted
          with a polynomial.
          Default to 2.

    Keyword Args:
        DPStep (Optional[float]):       Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        mcf (float/array):    Momentum compaction factor :math:`\alpha`
          up to the order fit_order. Returns a float if fit_order=1 otherwise
          returns an array.
    """
    if n_step < 2*fit_order:
        raise ValueError('Low nunber of steps, it is advised to have n_step >= 2*fit_order'+
        ' for a better fit.')
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    dp_samples = numpy.linspace(-dp_step/2, dp_step/2, n_step)
    fp_i = tuple(find_orbit4(ring, dp=dp + dp_i, keep_lattice=keep_lattice)[0] \
       for dp_i in dp_samples)
    fp = numpy.stack(fp_i, axis=0).T  # generate a Fortran contiguous array
    b = numpy.squeeze(internal_lpass(ring, fp, keep_lattice=True), axis=(2, 3))
    ring_length = get_s_pos(ring, len(ring))
    p = numpy.polynomial.Polynomial.fit(b[4], b[5], deg=fit_order).convert().coef
    alphac = p[1:] / ring_length[0]
    return alphac[0] if len(alphac)<2 else alphac

def get_slip_factor(ring: Lattice, **kwargs) -> float:
    r"""Compute the slip factor :math:`\eta`

    Parameters:
        ring:           Lattice description (:pycode:`ring.is_6d` must be
          :py:obj:`False`)

    Keyword Args:
        dp (float):     Momentum deviation
        DPStep (float): Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        eta (float):    Slip factor :math:`\eta`
    """
    gamma = ring.gamma
    etac = (1.0/gamma/gamma - get_mcf(ring, **kwargs))
    return etac


def get_revolution_frequency(ring: Lattice,
                             dp: float = None,
                             dct: float = None,
                             df: float = None) -> float:
    """Compute the revolution frequency of the full ring [Hz]

    Parameters:
        ring:       Lattice description
        dp:         Momentum deviation. Defaults to :py:obj:`None`
        dct:        Path lengthening. Defaults to :py:obj:`None`
        df:         Deviation of RF frequency. Defaults to :py:obj:`None`

    Returns:
        frev:       Revolution frequency [Hz]
    """
    lcell = ring.cell_length
    cell_frev = ring.beta * clight / lcell
    if dct is not None:
        cell_frev *= lcell / (lcell + dct)
    elif dp is not None:
        # Find the path lengthening for dp
        rnorad = ring.disable_6d(copy=True) if ring.is_6d else ring
        orbit = internal_lpass(rnorad, rnorad.find_orbit4(dp=dp)[0])
        dct = numpy.squeeze(orbit)[5]
        cell_frev *= lcell / (lcell + dct)
    elif df is not None:
        cell_frev += df / ring.cell_harmnumber
    return cell_frev / ring.periodicity


def set_rf_frequency(ring: Lattice, frequency: float = None,
                     dp: float = None, dct: float = None, df: float = None,
                     **kwargs):
    """Set the RF frequency

    Parameters:
        ring:           Lattice description
        frequency:      RF frequency [Hz]. Default: nominal frequency.
        dp:             Momentum deviation. Defaults to :py:obj:`None`
        dct:            Path lengthening. Defaults to :py:obj:`None`
        df:             Deviation of RF frequency. Defaults to :py:obj:`None`

    Keyword Args:
        cavpts (Optional[Refpts]):  If :py:obj:`None`, look for ring.cavpts, or
          otherwise take all cavities.
        array (Optional[bool]):     If :py:obj:`False` (default), *frequency*
          is applied to the selected cavities with the lowest frequency. The
          frequency of all the other selected cavities is scaled by the same
          ratio.

          If :py:obj:`True`, directly apply *frequency* to the selected
          cavities. The value must be broadcastable to the number of cavities.
        copy (Optional[bool]):     If :py:obj:`True`, returns a shallow copy of
          *ring* with new cavity elements. Otherwise (default), modify
          *ring* in-place
    """
    if frequency is None:
        frequency = get_revolution_frequency(ring, dp=dp, dct=dct, df=df) \
                    * ring.harmonic_number
    return ring.set_cavity(Frequency=frequency, **kwargs)


Lattice.mcf = property(get_mcf, doc="Momentum compaction factor")
Lattice.slip_factor = property(get_slip_factor, doc="Slip factor")
Lattice.get_mcf = get_mcf
Lattice.get_slip_factor = get_slip_factor
Lattice.get_revolution_frequency = get_revolution_frequency
Lattice.set_rf_frequency = set_rf_frequency
Lattice.rf_frequency = property(get_rf_frequency, set_rf_frequency,
    doc="Fundamental RF frequency [Hz]. The special value "
        ":py:class:`Frf.NOMINAL <.Frf>` means nominal frequency.")



import numpy as np
import at
import matplotlib.pyplot as plt
import pickle as pkl
from scipy.optimize import curve_fit


def get_growth_rates():
    '''
    Function to compute all LCBI mode growth rates
    Largely stolen from mbtrack2 (Alexis Gamelin)
    '''

    def Zr(omega):
        df = (omega_res/omega - omega/omega_res)
        return np.real(rshunt / (1 + 1j * qfactor * df))

    factor = eta * current / (4 * np.pi * ring.energy * qs)

    growth_rates = np.zeros(Nbunches)
    for mu in np.arange(Nbunches):
        omega0 = 2 * np.pi * f0
        omega_res = 2 * np.pi * fres
        n_max = int(10 * omega_res / (omega0 * Nbunches))

        n0 = np.arange(n_max)
        n1 = np.arange(1, n_max)
        omega_p = omega0 * (n0 * Nbunches + mu + qs)
        omega_m = omega0 * (n1 * Nbunches - mu - qs)

        sum_val = np.sum(omega_p * Zr(omega_p)) - np.sum(omega_m * Zr(omega_m))
        growth_rates[mu] = factor * sum_val
    return growth_rates


def sin_fit(x, amp, freq, phase, off):
    return amp * np.sin(freq * x + phase) + off


def getTimeVectors(tbtMatrix):

    print('Do SVD')
    u, s, v = np.linalg.svd(tbtMatrix)

    # time vector of strongest mode
    vector = np.array([v[0][k] for k in np.arange(v.shape[1])])

    # SVD does not give mode number, so we fit a sin to the spatial pattern
    fit, _ = curve_fit(sin_fit, np.arange(Nbunches), np.real(u[:, 0]),
                       p0=[np.amax(np.real(u[:, 0])),
                       2*np.pi*expected_mode/Nbunches,
                       -np.pi/4, 0])
    mode = fit[1]/2/np.pi*Nbunches

    bunchPlot = np.arange(0, Nbunches, 0.01)

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(np.real(u[:, 0]),
             color='r', marker='x', linestyle='None', label='Spatial Pattern')
    ax1.plot(bunchPlot, sin_fit(bunchPlot, fit[0], fit[1], fit[2], fit[3]),
             color='b', linestyle='solid', label='Sinusoidal Fit')
    ax1.set_xlabel('Bunch')
    ax1.legend()
    plt.show()

    return vector, mode, fit


def fitTimeVectors(tv):

    width = 400
    step = 200

    numsteps = int((len(tv) - width)/step) + 1
    starts = np.arange(0, len(tv) - width + 1, step)
    ends = np.arange(width, len(tv) + 1, step)

    dat = np.real(np.abs(tv))

    newdat = np.array([np.amax(dat[starts[i]:ends[i]])
                       for i in np.arange(numsteps)])

    turns = np.arange(len(dat))
    fitVal = np.polyfit((starts+ends) / 2, np.log(newdat), 1)
    chi_squared = np.sum((
                  np.polyval(fitVal, (starts+ends)/2) - np.log(newdat))**2)

    if 1/fitVal[0] > 1e6:
        fitVal[0] = np.inf

    return fitVal, newdat, starts, ends, chi_squared


# First we define everything we need and get the theoretical growth rates
resDict = pkl.load(open('./LCBI_output.pkl', 'rb'))
ring = resDict['ring']
Nbunches = resDict['M']
current = resDict['current']
fres = resDict['fres']
qfactor = resDict['qfactor']
rshunt = resDict['rshunt']


f0 = ring.revolution_frequency
alpha = at.get_mcf(ring.radiation_off(copy=True))
eta = at.get_slip_factor(ring.radiation_off(copy=True))
envel = at.envelope_parameters(ring.radiation_on(copy=True))
qx, qy, qs = ring.get_tune()
betaz = ring.circumference * alpha / (2 * np.pi * qs)
fs0 = qs*f0

gr_theory = get_growth_rates()
expected_mode = np.argmax(gr_theory)
gr_peak = gr_theory[expected_mode]


data_dp = resDict['dp'].T
data_z = resDict['z'].T

# Plot all data dp, just to see how it looks
bbs = np.arange(0, Nbunches)
for bb in bbs:
    plt.plot(data_dp[bb, :])
plt.xlabel('Turn')
plt.ylabel('dp/p')
plt.show()

# Here we use dp and z and combine with betaz into one complex number
# this allows us to distinguish from + and - modes
inv = np.array([[1/np.sqrt(betaz), 0],
                [0, np.sqrt(betaz)]])

data = np.zeros(data_dp[:, :].shape) + 1j * 0
for i in np.arange(data_dp.shape[0]):
    dp = data_dp[i, :]
    z = data_z[i, :]
    datpos, datmom = np.matmul(inv, [z, dp])
    data[i, :] = datpos + 1j * datmom

# Do the SVD and get all the time vectors
tv, mode, _ = getTimeVectors(data)
mode = int(np.round(mode))

# Fit the time vector
gr, newdat, starts, ends, chis = fitTimeVectors(tv)

turns = np.arange(0, tv.shape[0], 50)

fig = plt.figure()
ax1 = fig.add_subplot(111)
ax1.plot(np.real(tv), color='r', linestyle='solid', marker='None')
ax1.plot(turns, np.exp(gr[1])*np.exp(turns*gr[0]),
         linestyle='solid', marker='None',
         label=r'tau={:.2f} turns'.format(1/gr[0]))
ax1.plot((starts + ends)/2, newdat, color='k', marker='x', linestyle='None')
ax1.legend()
ax1.set_title('Mode {:d}, mu={:.3f}'.format(mode, mode/Nbunches))
plt.show()


gr_pyat = gr[0] * f0  # growth rate in seconds

fig = plt.figure()
ax1 = fig.add_subplot(111)
ax1.plot([mode], [gr_pyat],
         color='r', marker='x', linestyle='None', label='PyAT')
ax1.plot(gr_theory, color='b', marker='o', linestyle='None', label='Theory')
ax1.set_xlabel('Mode')
ax1.set_ylabel('Growth Rate [1/s]')
ax1.legend()
plt.show()



import at
import numpy
import pytest
import warnings
from numpy.testing import assert_allclose as assert_close
from at.collective import Wake, WakeElement, ResonatorElement
from at.collective import WakeComponent, ResWallElement
from at.collective import add_beamloading, remove_beamloading, BLMode
from at import lattice_track
from at import lattice_pass, internal_lpass


_issorted = lambda a: numpy.all(a[:-1] <= a[1:])


def test_fillpattern(hmba_lattice):
    ring = hmba_lattice.enable_6d(copy=True)
    nbunch = 16
    current = 0.2
    ring.set_fillpattern(nbunch)
    ring.beam_current = current
    assert ring.fillpattern.shape == (ring.harmonic_number, )
    assert numpy.sum(ring.fillpattern) == 1
    assert ring.bunch_list.shape == (nbunch, )
    assert ring.nbunch == nbunch
    assert ring.bunch_currents.shape == (nbunch, )
    assert numpy.sum(ring.bunch_currents) == current
    

def test_build_srange():
    bspacing = 10
    wlength = 20
    bsize = 0.1
    steps = 1.0e-3
    stepl = 1.0e-2
    srange = Wake.build_srange(-bsize, bsize, steps, stepl, bspacing, wlength)
    assert len(numpy.unique(srange)) == len(srange)
    assert _issorted(srange)
    assert_close(max(srange), wlength + bsize - bspacing - stepl)
    assert min(srange) == -bsize
    assert_close(min(numpy.diff(srange)) ,steps)
    assert_close(min(numpy.diff(srange[srange>bsize])) ,stepl)
    
                  
def test_wake_object():
    srange = [0, 1, 1]
    with pytest.raises(Exception):
        Wake(srange)
    srange = Wake.build_srange(-0.36, 0.36, 1.0e-5, 1.0e-2, 844, 844)
    with pytest.raises(Exception):
        Wake.resistive_wall(srange, Wake.WakeComponent.X, 1.0, 1.0e-3, 1.0)
    srange = Wake.build_srange(0.0, 0.36, 1.0e-5, 1.0e-2, 844, 844) 
    long_res =  Wake.long_resonator(srange, 1.0e9, 1.0, 1.0e3, 1.0)
    assert 1.0e-24 in long_res.srange
    assert long_res.Z.shape == long_res.srange.shape
    
    
def test_wake_element(hmba_lattice):
    ring = hmba_lattice.enable_6d(copy=True)
    srange = Wake.build_srange(0.0, 0.36, 1.0e-5, 1.0e-2, 844, 844) 
    long_res =  Wake.long_resonator(srange, 1.0e9, 1.0, 1.0e3, 1.0) 
    welem = WakeElement('WELEM', ring, long_res)
    assert welem.TurnHistory.shape == (ring.nbunch*welem.Nslice*welem.Nturns, 4)
    ring.append(welem)
    ring.set_fillpattern(16)
    assert welem.TurnHistory.shape == (ring.nbunch*welem.Nslice*welem.Nturns, 4)
    ring.disable_6d(at.Collective)
    assert welem.PassMethod == 'IdentityPass'
    
    
def test_resonator_element(hmba_lattice):
    ring = hmba_lattice.enable_6d(copy=True)
    srange = Wake.build_srange(0.0, 0.36, 1.0e-5, 1.0e-2, 844, 844)
    welem = ResonatorElement('WELEM', ring, srange, WakeComponent.Z, 1.0e9, 1.0, 1.0e3)  
    assert welem.ResFrequency == 1.0e9
    assert welem.Qfactor == 1
    assert welem.Rshunt == 1.0e3
    assert_close(welem.WakeZ[[0,1]], [3.14159265e+12, 6.28318531e+12], rtol = 1e-8)
    welem.ResFrequency += 100
    assert welem.ResFrequency == 1.0e9 + 100
    
    
def test_resistive_wall_element(hmba_lattice):
    ring = hmba_lattice.enable_6d(copy=True)
    srange = Wake.build_srange(0.0, 0.36, 1.0e-5, 1.0e-2, 844, 844)
    welem = ResWallElement('WELEM', ring, srange, WakeComponent.DX, 1.0, 1.0e-2, 1.0e6)  
    assert welem.RWLength == 1.0
    assert welem.Conductivity == 1.0e6
    assert_close(welem.WakeDX[[0,1]], [ 0.0000000e+00, -1.0449877e+24], rtol = 1e-8)
    welem.Conductivity += 100
    assert welem.Conductivity == 1.0e6 + 100
    
    
def test_beamloading(hmba_lattice):
    ring = hmba_lattice.enable_6d(copy=True)
    with pytest.raises(Exception):
        add_beamloading(ring, 44e3, 400, cavpts=range(len(ring)))
    add_beamloading(ring, 44e3, 400)
    cavs = ring.get_elements(at.RFCavity)  
    for cav in cavs:
        assert cav.PassMethod == 'BeamLoadingCavityPass'
        assert hasattr(cav, 'Vbeam') 
        assert hasattr(cav, 'Vgen')
        assert hasattr(cav, 'Vcav') 
    ring.disable_6d(at.RFCavity)
    for cav in cavs:
        assert cav.PassMethod == 'IdentityPass'  
    ring.enable_6d(at.RFCavity)
    for cav in cavs:
        assert cav.PassMethod == 'BeamLoadingCavityPass'          
    remove_beamloading(ring)
    cavs = ring.get_elements(at.RFCavity)  
    for cav in cavs:
        assert cav.PassMethod == 'RFCavityPass' 
    

@pytest.mark.parametrize('func', (lattice_track, lattice_pass))
def test_track_beamloading(hmba_lattice, func):
    ring = hmba_lattice.enable_6d(copy=True)
    rin0 = numpy.zeros(6)
    func(ring, rin0, refpts=[])
    add_beamloading(ring, 44e3, 400, blmode=BLMode.WAKE)
    rin1 = numpy.zeros(6)
    func(ring, rin1, refpts=[])
    assert_close(rin0, rin1, atol=1e-21)
    ring.set_fillpattern(2)
    ring.beam_current = 0.2
    rin = numpy.zeros((6, 1))
    with pytest.raises(Exception):
        func(ring, rin, refpts=[])
    rin = numpy.zeros((6, 2))
    if func == lattice_track:
        func(ring, rin, refpts=[], in_place=True)
    else:
        func(ring, rin, refpts=[])
    assert_close(rin[:, 0], numpy.array([-2.318948e-08, -1.599715e-09,
                                        0.000000e+00,  0.000000e+00,
                                        -1.313306e-05, -1.443748e-08]
                                        ), atol=5e-10)
 

def test_buffers(hmba_lattice):
    ring = hmba_lattice.enable_6d(copy=True)
    ring.periodicity = 1
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ring.harmonic_number = 32
    nturns = 11
    nbunch = 4
    nslice = 51
    ns = nbunch*nslice
    ls = ns*ring.circumference/ring.periodicity
    add_beamloading(ring, 44e3, 400, Nturns=nturns, Nslice=nslice,
                    buffersize=nturns, blmode=BLMode.WAKE)
    ring.set_fillpattern(nbunch)
    ring.beam_current = 0.2
    rin = numpy.zeros((6, nbunch)) + 1.0e-6
    bl_elem = ring.get_elements('*_BL')[0]
    th = numpy.zeros((nturns, ) + bl_elem.TurnHistory.shape)
    vbh = numpy.zeros((nturns, ) + bl_elem.Vbeam_buffer.shape)
    vgh = numpy.zeros((nturns, ) + bl_elem.Vgen_buffer.shape)
    vbbh = numpy.zeros((nturns, ) + bl_elem.Vbunch_buffer.shape)
    for i in numpy.arange(nturns):
        ring.track(rin, nturns=1, refpts=None, in_place=True)
        th[i] = bl_elem.TurnHistory
        vbh[i] = bl_elem.Vbeam_buffer
        vgh[i] = bl_elem.Vgen_buffer
        vbbh[i] = bl_elem.Vbunch_buffer
    for i in numpy.arange(1, nturns):
        dth = numpy.sum(th[i-1, (nturns-i)*ns:]
                        - th[i, (nturns-i-1)*ns:(nturns-1)*ns]) - i*ls
        dvbh = numpy.sum(vbh[i-1, :, (nturns-i):]
                         - vbh[i, :, (nturns-i-1):(nturns-1)])
        dvgh = numpy.sum(vgh[i-1, :, (nturns-i):]
                         - vgh[i, :, (nturns-i-1):(nturns-1)])
        dvbbh = numpy.sum(vbbh[i-1, :, (nturns-i):]
                          - vbbh[i, :, (nturns-i-1):(nturns-1)])
    assert_close([dth, dvbh, dvgh, dvbbh], numpy.zeros(4), atol=1e-9)



r"""
Definition of :py:class:`Variable <.VariableBase>` objects used in matching and
response matrices.

See :ref:`example-notebooks` for examples of matching and response matrices.

Each :py:class:`Variable <.VariableBase>` has a scalar value.

.. rubric:: Class hierarchy

:py:class:`VariableBase`\ (name, bounds, delta)

- :py:class:`~.lattice_variables.ElementVariable`\ (elements, attrname, index, ...)
- :py:class:`~.lattice_variables.RefptsVariable`\ (refpts, attrname, index, ...)
- :py:class:`CustomVariable`\ (setfun, getfun, ...)

.. rubric:: VariableBase methods

:py:class:`VariableBase` provides the following methods:

- :py:meth:`~VariableBase.get`
- :py:meth:`~VariableBase.set`
- :py:meth:`~VariableBase.set_previous`
- :py:meth:`~VariableBase.reset`
- :py:meth:`~VariableBase.increment`
- :py:meth:`~VariableBase.step_up`
- :py:meth:`~VariableBase.step_down`

.. rubric:: VariableBase properties

:py:class:`.VariableBase` provides the following properties:

- :py:attr:`~VariableBase.initial_value`
- :py:attr:`~VariableBase.last_value`
- :py:attr:`~VariableBase.previous_value`
- :py:attr:`~VariableBase.history`

The :py:class:`VariableBase` abstract class may be used as a base class to define
custom variables (see examples). Typically, this consist in overloading the abstract
methods *_setfun* and *_getfun*

.. rubric:: Examples

Write a subclass of :py:class:`VariableBase` which varies two drift lengths so
that their sum is constant:

.. code-block:: python

    class ElementShifter(at.VariableBase):
        '''Varies the length of the elements identified by *ref1* and *ref2*
        keeping the sum of their lengths equal to *total_length*.

        If *total_length* is None, it is set to the initial total length
        '''
        def __init__(self, drift1, drift2, total_length=None, **kwargs):
            # store the 2 variable elements
            self.drift1 = drift1
            self.drift2 = drift2
            # store the initial total length
            if total_length is None:
                total_length = drift1.Length + drift2.Length
            self.length = total_length
            super().__init__(bounds=(0.0, total_length), **kwargs)

        def _setfun(self, value, **kwargs):
            self.drift1.Length = value
            self.drift2.Length = self.length - value

        def _getfun(self, **kwargs):
            return self.drift1.Length

And create a variable varying the length of drifts *DR_01* and *DR_01* and
keeping their sum constant:

.. code-block:: python

    drift1 = hmba_lattice["DR_01"]
    drift2 = hmba_lattice["DR_02"]
    var2 = ElementShifter(drift1, drift2, name="DR_01")

"""

from __future__ import annotations

__all__ = [
    "VariableBase",
    "CustomVariable",
    "VariableList",
]

import abc
from collections import deque
from collections.abc import Iterable, Sequence, Callable
from typing import Union

import numpy as np

Number = Union[int, float]


def _nop(value):
    return value


class VariableBase(abc.ABC):
    """A Variable abstract base class

    Derived classes must implement the :py:meth:`~VariableBase._getfun` and
    :py:meth:`~VariableBase._getfun` methods
    """

    _counter = 0
    _prefix = "var"

    def __init__(
        self,
        *,
        name: str = "",
        bounds: tuple[Number, Number] = (-np.inf, np.inf),
        delta: Number = 1.0,
        history_length: int = None,
        ring=None,
    ):
        """
        Parameters:
            name:       Name of the Variable
            bounds:     Lower and upper bounds of the variable value
            delta:      Initial variation step
            history_length: Maximum length of the history buffer. :py:obj:`None`
              means infinite
            ring:       provided to an attempt to get the initial value of the
              variable
        """
        self.name: str = self._setname(name)  #: Variable name
        self.bounds: tuple[Number, Number] = bounds  #: Variable bounds
        self.delta: Number = delta  #: Increment step
        #: Maximum length of the history buffer. :py:obj:`None` means infinite
        self.history_length = history_length
        self._initial = np.nan
        self._history = deque([], self.history_length)
        try:
            self.get(ring=ring, initial=True)
        except ValueError:
            pass

    @classmethod
    def _setname(cls, name):
        cls._counter += 1
        if name:
            return name
        else:
            return f"{cls._prefix}{cls._counter}"

    # noinspection PyUnusedLocal
    def _setfun(self, value: Number, ring=None):
        classname = self.__class__.__name__
        raise TypeError(f"{classname!r} is read-only")

    @abc.abstractmethod
    def _getfun(self, ring=None) -> Number: ...

    @property
    def history(self) -> list[Number]:
        """History of the values of the variable"""
        return list(self._history)

    @property
    def initial_value(self) -> Number:
        """Initial value of the variable"""
        if not np.isnan(self._initial):
            return self._initial
        else:
            raise IndexError(f"{self.name}: No value has been set yet")

    @property
    def last_value(self) -> Number:
        """Last value of the variable"""
        try:
            return self._history[-1]
        except IndexError as exc:
            exc.args = (f"{self.name}: No value has been set yet",)
            raise

    @property
    def previous_value(self) -> Number:
        """Value before the last one"""
        try:
            return self._history[-2]
        except IndexError as exc:
            exc.args = (f"{self.name}: history too short",)
            raise

    def set(self, value: Number, ring=None) -> None:
        """Set the variable value

        Args:
            value:  New value to be applied on the variable
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to set the variable.
        """
        if value < self.bounds[0] or value > self.bounds[1]:
            raise ValueError(f"set value must be in {self.bounds}")
        self._setfun(value, ring=ring)
        if np.isnan(self._initial):
            self._initial = value
        self._history.append(value)

    def get(
        self, ring=None, *, initial: bool = False, check_bounds: bool = False
    ) -> Number:
        """Get the actual variable value

        Args:
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to get the variable value.
            initial:    If :py:obj:`True`, clear the history and set the variable
              initial value
            check_bounds: If :py:obj:`True`, raise a ValueError if the value is out
              of bounds

        Returns:
            value:      Value of the variable
        """
        value = self._getfun(ring=ring)
        if initial or np.isnan(self._initial):
            self._initial = value
            self._history = deque([value], self.history_length)
        if check_bounds:
            if value < self.bounds[0] or value > self.bounds[1]:
                raise ValueError(f"value out of {self.bounds}")
        return value

    value = property(get, set, doc="Actual value")

    @property
    def _safe_value(self):
        try:
            v = self._history[-1]
        except IndexError:
            v = np.nan
        return v

    def set_previous(self, ring=None) -> None:
        """Reset to the value before the last one

        Args:
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to set the variable.
        """
        if len(self._history) >= 2:
            self._history.pop()  # Remove the last value
            value = self._history.pop()  # retrieve the previous value
            self.set(value, ring=ring)
        else:
            raise IndexError(f"{self.name}: history too short")

    def reset(self, ring=None) -> None:
        """Reset to the initial value and clear the history buffer

        Args:
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to reset the variable.
        """
        iniv = self._initial
        if not np.isnan(iniv):
            self._history = deque([], self.history_length)
            self.set(iniv, ring=ring)
        else:
            raise IndexError(f"reset {self.name}: No value has been set yet")

    def increment(self, incr: Number, ring=None) -> None:
        """Increment the variable value

        Args:
            incr:   Increment value
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to increment the variable.
        """
        if self._initial is None:
            self.get(ring=ring, initial=True)
        self.set(self.last_value + incr, ring=ring)

    def _step(self, step: Number, ring=None) -> None:
        if self._initial is None:
            self.get(ring=ring, initial=True)
        self.set(self._initial + step, ring=ring)

    def step_up(self, ring=None) -> None:
        """Set to initial_value + delta

        Args:
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to set the variable.
        """
        self._step(self.delta, ring=ring)

    def step_down(self, ring=None) -> None:
        """Set to initial_value - delta

        Args:
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to set the variable.
        """
        self._step(-self.delta, ring=ring)

    @staticmethod
    def _header():
        return "\n{:>12s}{:>13s}{:>16s}{:>16s}\n".format(
            "Name", "Initial", "Final ", "Variation"
        )

    def _line(self):
        vnow = self._safe_value
        vini = self._initial

        return "{:>12s}{: 16e}{: 16e}{: 16e}".format(
            self.name, vini, vnow, (vnow - vini)
        )

    def status(self):
        """Return a string describing the current status of the variable

        Returns:
            status: Variable description
        """
        return "\n".join((self._header(), self._line()))

    def __float__(self):
        return float(self._safe_value)

    def __int__(self):
        return int(self._safe_value)

    def __str__(self):
        return f"{self.__class__.__name__}({self._safe_value}, name={self.name!r})"

    def __repr__(self):
        return repr(self._safe_value)


class CustomVariable(VariableBase):
    r"""A Variable with user-defined get and set functions

    This is a convenience function allowing user-defined *get* and *set*
    functions. But subclassing :py:class:`.Variable` should always be preferred
    for clarity and efficiency.

    """

    def __init__(
        self,
        setfun: Callable,
        getfun: Callable,
        *args,
        name: str = "",
        bounds: tuple[Number, Number] = (-np.inf, np.inf),
        delta: Number = 1.0,
        history_length: int = None,
        ring=None,
        **kwargs,
    ):
        """
        Parameters:
            getfun:     Function for getting the variable value. Called as
              :pycode:`getfun(*args, ring=ring, **kwargs) -> Number`
            setfun:     Function for setting the variable value. Called as
              :pycode:`setfun(value: Number, *args, ring=ring, **kwargs): None`
            name:       Name of the Variable
            bounds:     Lower and upper bounds of the variable value
            delta:      Initial variation step
            *args:      Variable argument list transmitted to both the *getfun*
              and *setfun* functions. Such arguments can always be avoided by
              using :py:func:`~functools.partial` or callable class objects.

        Keyword Args:
            **kwargs:   Keyword arguments transmitted to both the *getfun*
              and *setfun* functions. Such arguments can always be avoided by
              using :py:func:`~functools.partial` or callable class objects.
        """
        self.getfun = getfun
        self.setfun = setfun
        self.args = args
        self.kwargs = kwargs
        super().__init__(
            name=name,
            bounds=bounds,
            delta=delta,
            history_length=history_length,
            ring=ring,
        )

    def _getfun(self, ring=None) -> Number:
        return self.getfun(*self.args, ring=ring, **self.kwargs)

    def _setfun(self, value: Number, ring=None):
        self.setfun(value, *self.args, ring=ring, **self.kwargs)


class VariableList(list):
    """Container for Variable objects

    :py:class:`VariableList` supports all :py:class:`list` methods, like
    appending, insertion or concatenation with the "+" operator.
    """

    def get(self, ring=None, **kwargs) -> Sequence[float]:
        r"""Get the current values of Variables

        Args:
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to set the variable.

        Keyword Args:
            initial:    If :py:obj:`True`, set the Variables'
              initial value
            check_bounds: If :py:obj:`True`, raise a ValueError if the value is out
              of bounds

        Returns:
            values:     1D array of values of all variables
        """
        return np.array([var.get(ring=ring, **kwargs) for var in self])

    def set(self, values: Iterable[float], ring=None) -> None:
        r"""Set the values of Variables

        Args:
            values:     Iterable of values
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to set the variable.
        """
        for var, val in zip(self, values):
            var.set(val, ring=ring)

    def increment(self, increment: Iterable[float], ring=None) -> None:
        r"""Increment the values of Variables

        Args:
            increment:  Iterable of values
            ring:   Depending on the variable type, a :py:class:`.Lattice` argument
              may be necessary to increment the variable.
        """
        for var, incr in zip(self, increment):
            var.increment(incr, ring=ring)

    # noinspection PyProtectedMember
    def status(self, **kwargs) -> str:
        """String description of the variables"""
        values = "\n".join(var._line(**kwargs) for var in self)
        return "\n".join((VariableBase._header(), values))

    def __str__(self) -> str:
        return self.status()

    @property
    def deltas(self) -> Sequence[Number]:
        """delta values of the variables"""
        return np.array([var.delta for var in self])



import itertools
import pytest

from at.load.elegant import (
    expand_elegant,
    parse_chunk,
    parse_lines,
    elegant_element_from_string,
)
from at.lattice.elements import (
    Dipole,
    Drift,
    Marker,
    Quadrupole,
    RFCavity,
    Sextupole,
)


@pytest.fixture
def defaults():
    return {"energy": 3e9, "harmonic_number": 936}


@pytest.mark.parametrize(
    "line,target",
    [
        ["a\n!b\nc", ["a", "c"]],
        ["a\nb!comment\nc", ["a", "b", "c"]],
        ["a\nb& !comment\nc\nd", ["a", "bc", "d"]],
        ["a&\nb", ["ab"]],
        ["a&\nb\nc", ["ab", "c"]],
    ],
)
def test_parse_lines(line, target):
    assert parse_lines(line) == target


def test_elegant_element_from_string_handles_drift(defaults):
    drift = "drift,l=0.0450000"
    assert elegant_element_from_string("d1", drift, defaults).equals(
        Drift("d1", 0.0450000)
    )


def test_elegant_element_from_string_handles_arithmetic(defaults):
    drift = 'drift,l="0.04 2 /"'
    assert elegant_element_from_string("d1",
                                       drift,
                                       defaults).equals(Drift("d1", 0.02))


def test_elegant_element_from_string_handles_marker(defaults):
    marker = "mark"
    expected = Marker("m1")
    assert elegant_element_from_string("m1", marker, defaults).equals(expected)


def test_elegant_element_from_string_handles_quadrupole(defaults):
    quad = "KQUAD, N_KICKS=30, L=0.4064, K1=-0.7008"
    quad = quad.lower()
    expected = Quadrupole(
        "q1", 0.4064, k=-0.7008, NumIntSteps=30,
        PassMethod="StrMPoleSymplectic4Pass",
    )
    assert elegant_element_from_string("q1", quad, defaults).equals(expected)


def test_elegant_element_from_string_handles_sextupole(defaults):
    sext = "KSEXT,N_KICKS=12, L=0.29, K2=39.55"
    sext = sext.lower()
    # Note that h is divided by two due to Elegant's scaling.
    expected = Sextupole("s1", 0.29, h=19.775, NumIntSteps=12)
    assert elegant_element_from_string("s1", sext, defaults).equals(expected)


def test_elegant_element_from_string_handles_cavity(defaults):
    cavity = "RFCA, L=0.0, VOLT=2.5e6, FREQ=499654000, PHASE=156.7"
    cavity = cavity.lower()
    expected = RFCavity("c1", 0.0, 2.5e6, 4.99654e8, 936, 3e9, Phi="156.7")
    constructed = elegant_element_from_string("c1", cavity, defaults)
    assert constructed.equals(expected)


def test_elegant_element_from_string_handles_bending(defaults):
    bend = "CSBEN,L=0.933,K1=0,Angle=0.1308,E1=0.06544,E2=0.06544," \
           "N_KICKS=50, HGAP=0.0233, FINT=0.6438"
    bend = bend.lower()
    expected = Dipole(
        "b1",
        0.933,
        BendingAngle=0.1308,
        EntranceAngle=0.06544,
        ExitAngle=0.06544,
        FringeInt1=0.6438,
        FringeInt2=0.6438,
        FullGap=0.0466,
        k=0,
        NumIntSteps=50,
        PassMethod="BndMPoleSymplectic4Pass",
        PolynomA=[0, 0, 0, 0, 0],
        PolynomB=[0, 0, 0, 0, 0],
    )
    assert elegant_element_from_string("b1", bend, defaults).equals(expected)


def test_elegant_element_from_string_handles_variable(defaults):
    drift = "drift,l=a"
    defaults["a"] = 1
    correct_drift = Drift("d1", 1)
    constructed_drift = elegant_element_from_string("d1", drift, defaults)
    assert correct_drift.equals(constructed_drift)


@pytest.mark.parametrize(
    "value,elements,chunks,expected",
    [
        ["line=(a)", {"a": Marker("m")}, {}, [Marker("m")]],
        ["2*a", {"a": Marker("m")}, {}, [Marker("m"), Marker("m")]],
        ["2*(a)", {"a": Marker("m")}, {}, [Marker("m"), Marker("m")]],
        ["-a", {"a": Marker("m")}, {}, [Marker("m")]],
        ["b", {}, {"b": [Marker("1"), Marker("2")]},
         [Marker("1"), Marker("2")]],
        ["-b", {}, {"b": [Marker("1"), Marker("2")]},
         [Marker("2"), Marker("1")]],
        [
            "-bb",
            {},
            {"bb": [Dipole("bb", 1.0, 0.05, EntranceAngle=0.02,
             ExitAngle=0.03,)]},
            [Dipole("bb", 1.0, 0.05, EntranceAngle=0.03, ExitAngle=0.02)],
        ],
    ],
)
def test_parse_chunk(value, elements, chunks, expected):
    result = parse_chunk(value, elements, chunks)
    try:
        zipped = itertools.zip_longest(result, expected)
    except AttributeError:
        # Python 2 compatibility.
        zipped = itertools.izip_longest(result, expected)

    for r, e in zipped:
        assert r.equals(e)


def test_expand_elegant():
    contents = """dmult:drift,l=1
diad6d:line=(dmult)"""
    elements = expand_elegant(contents, "diad6d", 3e9, 936)
    assert len(elements) == 1
    assert elements[0].equals(Drift("dmult", 1.0))



import matlab
import numpy as np
import pytest
from numpy.testing import assert_allclose as assert_close

import at


@pytest.mark.parametrize(
    "lattices",
    [
        "dba",
        "hmba",
        "hmba_cav",
        "hmba_rad",
    ],
)
def test_atpass(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    xy_step = 1.0e-8
    scaling = [xy_step, xy_step, xy_step, xy_step, xy_step, xy_step]

    ml_rin = engine.diag(matlab.double(scaling))
    ml_rout = engine.atpass(ml_lattice, ml_rin, 1, 1)

    py_rin = np.asfortranarray(np.diag(scaling))
    at.lattice_track(py_lattice, py_rin, 1, in_place=True)

    assert_close(py_rin, ml_rout, rtol=0, atol=1.0e-30)


@pytest.mark.parametrize(
    "lattices",
    [
        "dba",
        "hmba",
        "hmba_cav",
        "hmba_rad",
    ],
)
def test_linepass(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    xy_step = 1.0e-8
    scaling = [xy_step, xy_step, xy_step, xy_step, xy_step, xy_step]

    ml_rin = engine.diag(matlab.double(scaling))
    ml_rout = np.array(engine.linepass(ml_lattice, ml_rin))

    py_rin = np.asfortranarray(np.diag(scaling))
    py_rout, *_ = at.lattice_track(py_lattice, py_rin, refpts=len(py_lattice))
    py_rout = np.squeeze(py_rout)

    assert_close(py_rout, ml_rout, rtol=0, atol=1.0e-30)



"""Simple tests to make sure integrators load in Python."""
import numpy
import pytest

# noinspection PyUnresolvedReferences,PyProtectedMember
from at.lattice import Element, elements
from at import shift_elem, tilt_elem
from at import element_track, lattice_track
from at import lattice_pass, internal_lpass
from at import element_pass, internal_epass


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_exact_hamiltonian_pass(rin, func):
    drift = elements.Multipole('m1', 1, [0, 0, 0, 0], [0, 0, 0, 0])
    drift.Type = 0
    drift.PassMethod = 'ExactHamiltonianPass'
    drift.BendingAngle = 0
    func(drift, rin)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_exact_hamiltonian_pass_with_dls_dipole(rin, func):
    bend = elements.Multipole('rb', 0.15, [0, 0, 0, 0],
                              [-0.0116333, 3.786786, 0, 0])
    bend.Type = 1
    bend.PassMethod = 'ExactHamiltonianPass'
    bend.BendingAngle = -0.001745
    bend.Energy = 3.5e9
    bend.MaxOrder = 3
    if func==element_track:
        func(bend, rin, in_place=True)
    else:
        func(bend, rin)
    # Results from Matlab
    expected = numpy.array([9.23965e-9, 1.22319e-5, 0,
                            0, 0, -4.8100e-10]).reshape(6, 1)
    numpy.testing.assert_allclose(rin, expected, rtol=1e-5, atol=1e-6)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
@pytest.mark.parametrize('passmethod',
                         ('GWigSymplecticPass', 'GWigSymplecticRadPass'))
def test_gwig_symplectic_pass(rin, passmethod, func):
    # Parameters copied from one of the Diamond wigglers.
    wiggler = elements.Wiggler('w', 1.15, 0.05, 0.8, 3e9)
    wiggler.PassMethod = passmethod
    func(wiggler, rin)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_bndstrmpole_symplectic_4_pass(rin, func):
    bend = elements.Dipole('b', 1.0)
    bend.PassMethod = 'BndStrMPoleSymplectic4Pass'
    func(bend, rin)


@pytest.mark.parametrize('func', (element_track, element_pass, internal_epass))
def test_pydrift(func):
    pydrift = elements.Drift('drift', 1.0, PassMethod='pyDriftPass')
    cdrift = elements.Drift('drift', 1.0, PassMethod='DriftPass')
    pyout, *_ = func(pydrift, numpy.zeros(6)+1.0e-6)
    cout, *_ = func(cdrift, numpy.zeros(6)+1.0e-6)
    numpy.testing.assert_equal(pyout, cout)

    shift_elem(pydrift, 1.0e-3, 1.0e-3)
    shift_elem(cdrift, 1.0e-3, 1.0e-3)
    pyout, *_ = func(pydrift, numpy.zeros(6) + 1.0e-6)
    cout, *_ = func(cdrift, numpy.zeros(6) + 1.0e-6)
    numpy.testing.assert_equal(pyout, cout)

    tilt_elem(pydrift, 1.0e-3, 1.0e-3)
    tilt_elem(cdrift, 1.0e-3, 1.0e-3)
    pyout, *_ = func(pydrift, numpy.zeros(6)+1.0e-6)
    cout, *_ = func(cdrift, numpy.zeros(6)+1.0e-6)
    numpy.testing.assert_equal(pyout, cout)

    # Multiple particles
    pyout, *_ = func(pydrift, numpy.zeros((6, 2))+1.0e-6)
    cout, *_ = func(cdrift, numpy.zeros((6, 2))+1.0e-6)
    numpy.testing.assert_equal(pyout, cout)


@pytest.mark.parametrize('func', (lattice_track, lattice_pass, internal_lpass))
def test_pyintegrator(hmba_lattice, func):
    params = {'Length': 0,
              'PassMethod': 'pyIdentityPass',
              }
    id_elem = Element('py_id', **params)
    pin = numpy.zeros((6, 2))+1.0e-6
    pout1, *_ = func(hmba_lattice, pin.copy(), nturns=1)
    hmba_lattice = hmba_lattice + [id_elem]
    pout2, *_ = func(hmba_lattice, pin.copy(), nturns=1)
    numpy.testing.assert_equal(pout1, pout2)



"""
Helper functions for working with AT lattices.

A :py:class:`.Lattice` in pyAT is a sequence of :py:class:`.Element` objects.
These functions are useful for building ad manipulating these sequences.
"""
import sys
import numpy as np
from .axisdef import *
from .options import DConstant, random
from .particle_object import Particle
# from .variables import *
from .variables import VariableList
from .elements import *
from .rectangular_bend import *
from .idtable_element import InsertionDeviceKickMap
from .utils import *
from .lattice_object import *
# from .lattice_variables import *
from .cavity_access import *
from .variable_elements import *
from .deprecated import *
# Define the module "lattice.constants" for backward compatibility
# noinspection PyUnresolvedReferences
from .. import constants
sys.modules['at.lattice.constants'] = sys.modules['at.constants']

# Type definitions
Orbit = np.ndarray



from .lattice_object import Lattice
from .elements import Element
from .utils import Refpts, BoolRefpts, Uint32Refpts
from .utils import checkattr, get_uint32_index, get_bool_index
from typing import Sequence

__all__ = ['get_cells', 'get_refpts']


# noinspection PyIncorrectDocstring
def get_cells(ring: Sequence[Element], refpts: Refpts, *args,
              regex=False) -> BoolRefpts:
    # noinspection PyShadowingNames
    r"""
    get_cells(ring, filtfunc) -> BoolRefpts
    get_cells(ring, element_type) -> BoolRefpts
    get_cells(ring, attrname) -> BoolRefpts
    get_cells(ring, attrname, attrvalue) -> BoolRefpts
    Returns a bool array of element indices, selecting ring elements.

    Deprecated: :pycode:`get_cells(ring, refpts)` is
    :pycode:`ring.bool_refpts(refpts)` except for :py:obj:`str` arguments:
    :pycode:`get_cells(ring, attrname [, attrvalue])` is
    :pycode:`ring.bool_refpts(checkattr(strkey [, attrvalue]))`

    Parameters:
        ring (Sequence[Element]):       Lattice description
        filtfunc (ElementFilter):       Filter function. Selects
          :py:class:`.Element`\ s satisfying the filter function
        element_type (Type[Element]):   Element type
        attrname (str):                 Attribute name
        attrvalue (Any):                Attribute value. If absent, select the
          presence of an *attrname* attribute. If present, select
          :py:class:`.Element`\ s with :pycode:`attrname == attrvalue`.
        regex: Use regular expression for refpts string matching;
            Default: False (Unix shell-style wildcards)

    Returns:
        bool_refs (BoolRefpts):  numpy Array of :py:obj:`bool` with length
          len(ring)+1

    Examples:

        >>> refpts = get_cells(ring, 'Frequency')

        Returns a numpy array of booleans where all elements having a
        :pycode:`Frequency` attribute are :py:obj:`True`

        >>> refpts = get_cells(ring, 'K', 0.0)

        Returns a numpy array of booleans where all elements having a
        :pycode:`K` attribute equal to 0.0 are :py:obj:`True`

    See also:
        :py:meth:`.Lattice.bool_refpts`, :py:meth:`.Lattice.uint32_refpts`
    """
    if isinstance(refpts, str):
        refpts = checkattr(refpts, *args)
    return get_bool_index(ring, refpts, regex=regex)


# noinspection PyUnusedLocal,PyIncorrectDocstring
def get_refpts(ring: Sequence[Element], refpts: Refpts,
               regex=False) -> Uint32Refpts:
    r"""Return a :py:obj:`~numpy.uint32` array of element indices selecting
    ring elements.

    Deprecated: :pycode:`get_elements(ring, refpts)` is
    :pycode:`ring.uint32_refpts(refpts)`

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for refpts string matching;
            Default: False (Unix shell-style wildcards)

    Returns:
        uint32_refs (Uint32Refs):    :py:obj:`~numpy.uint32` numpy array as
          long as the number of refpts

    See also:
        :py:meth:`.Lattice.uint32_refpts`, :py:meth:`.Lattice.bool_refpts`
    """
    return get_uint32_index(ring, refpts, regex=regex)


Lattice.uint32_refpts = get_uint32_index
Lattice.bool_refpts = get_bool_index
Lattice.get_cells = get_cells
Lattice.get_refpts = get_refpts



import numpy as np
import pytest

from at.physics import get_radiation_integrals


@pytest.mark.parametrize("lattices", ["hmba"])
def test_radiation_integrals(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)

    # Python call
    py_integrals = get_radiation_integrals(py_lattice)
    # Matlab call
    results = engine.pyproxy("ringpara", ml_lattice)
    ml_integrals = np.squeeze(results["integrals"])
    # Comparison
    np.testing.assert_almost_equal(py_integrals, ml_integrals[:5])



"""
Accelerator physics functions
"""
# noinspection PyUnresolvedReferences
# Avalailable here for backward compatibility
from ..lattice import DConstant, Orbit, frequency_control
from .amat import *
from .energy_loss import *
from .revolution import *
from .harmonic_analysis import *
from .orbit import *
from .matrix import *
from .linear import *
from .diffmatrix import find_mpole_raddiff_matrix
from .radiation import *
from .ring_parameters import *
from .nonlinear import *
from .fastring import *
from .frequency_maps import fmap_parallel_track



"""
Closed orbit related functions
"""
import numpy
from at.constants import clight
from at.lattice import AtError, AtWarning, check_6d, DConstant, Orbit
from at.lattice import Lattice, get_s_pos, Refpts, frequency_control
from at.tracking import internal_lpass
from .energy_loss import ELossMethod, get_timelag_fromU0
import warnings

__all__ = ['find_orbit4', 'find_sync_orbit', 'find_orbit6', 'find_orbit']


@check_6d(False)
def _orbit_dp(ring: Lattice, dp: float = None, guess: Orbit = None, **kwargs):
    """Solver for fixed energy deviation"""
    # We seek
    #  - f(x) = x
    #  - g(x) = f(x) - x = 0
    #  - g'(x) = f'(x) - 1
    # Use a Newton-Raphson-type algorithm:
    #  - r_n+1 = r_n - g(r_n) / g'(r_n)
    #  - r_n+1 = r_n - (f(r_n) - r_n) / (f'(r_n) - 1)
    #
    # (f(r_n) - r_n) / (f'(r_n) - 1) can be seen as x = b/a where we use least
    #     squares fitting to determine x when ax = b
    # f(r_n) - r_n is denoted b
    # f'(r_n) is the 4x4 jacobian, denoted j4
    keep_lattice = kwargs.pop('keep_lattice', False)
    convergence = kwargs.pop('convergence', DConstant.OrbConvergence)
    max_iterations = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    kwargs.pop('DPStep', DConstant.DPStep)
    rem = kwargs.keys()
    if len(rem) > 0:
        raise AtError(f'Unexpected keywords for orbit_dp: {", ".join(rem)}')

    ref_in = numpy.zeros((6,)) if guess is None else numpy.copy(guess)
    ref_in[4] = 0.0 if dp is None else dp

    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0])
    delta_matrix = numpy.zeros((6, 5), order='F')
    for i in range(4):
        delta_matrix[i, i] = scaling[i]
    id4 = numpy.asfortranarray(numpy.identity(4))
    change = 1
    itercount = 0
    while (change > convergence) and itercount < max_iterations:
        in_mat = ref_in.reshape((6, 1)) + delta_matrix
        _ = internal_lpass(ring, in_mat, refpts=[], keep_lattice=keep_lattice)
        # the reference particle after one turn
        ref_out = in_mat[:, 4]
        # 4x4 jacobian matrix from numerical differentiation:
        # f(x+d) - f(x) / d
        j4 = (in_mat[:4, :4] - in_mat[:4, 4:]) / scaling
        a = j4 - id4  # f'(r_n) - 1
        b = ref_out[:4] - ref_in[:4]
        b_over_a = numpy.linalg.solve(a, b)
        r_next = ref_in - numpy.append(b_over_a, numpy.zeros((2,)))
        # determine if we are close enough
        change = numpy.linalg.norm(r_next - ref_in)
        itercount += 1
        ref_in = r_next
        keep_lattice = True

    if itercount == max_iterations:
        warnings.warn(AtWarning('Maximum number of iterations reached. '
                                'Possible non-convergence'))
    return ref_in


@check_6d(False)
def _orbit_dct(ring: Lattice, dct: float = None, guess: Orbit = None, **kwargs):
    """Solver for fixed path lengthening"""
    keep_lattice = kwargs.pop('keep_lattice', False)
    convergence = kwargs.pop('convergence', DConstant.OrbConvergence)
    max_iterations = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    kwargs.pop('DPStep', DConstant.DPStep)
    rem = kwargs.keys()
    if len(rem) > 0:
        raise AtError(f'Unexpected keywords for orbit_dct: {", ".join(rem)}')

    ref_in = numpy.zeros((6,)) if guess is None else numpy.copy(guess)

    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 1.0])
    delta_matrix = numpy.zeros((6, 6), order='F')
    for i in range(5):
        delta_matrix[i, i] = scaling[i]
    theta5 = numpy.zeros((5,), order='F')
    theta5[4] = 0.0 if dct is None else dct
    id5 = numpy.zeros((5, 5), order='F')
    for i in range(4):
        id5[i, i] = 1.0
    idx = numpy.array([0, 1, 2, 3, 5])
    change = 1
    itercount = 0
    while (change > convergence) and itercount < max_iterations:
        in_mat = ref_in.reshape((6, 1)) + delta_matrix
        _ = internal_lpass(ring, in_mat, refpts=[], keep_lattice=keep_lattice)
        # the reference particle after one turn
        ref_out = in_mat[:, -1]
        # 5x5 jacobian matrix from numerical differentiation:
        # f(x+d) - f(x) / d
        j5 = (in_mat[idx, :5] - in_mat[idx, 5:]) / scaling
        a = j5 - id5  # f'(r_n) - 1
        b = ref_out[idx] - numpy.append(ref_in[:4], 0.0) - theta5
        b_over_a = numpy.linalg.solve(a, b)
        r_next = ref_in - numpy.append(b_over_a, 0.0)
        # determine if we are close enough
        change = numpy.linalg.norm(r_next - ref_in)
        itercount += 1
        ref_in = r_next
        keep_lattice = True

    if itercount == max_iterations:
        warnings.warn(AtWarning('Maximum number of iterations reached. '
                                'Possible non-convergence'))
    return ref_in


def find_orbit4(ring: Lattice, dp: float = None, refpts: Refpts = None, *,
                dct: float = None,
                df: float = None,
                orbit: Orbit = None,
                keep_lattice: bool = False, **kwargs):
    r"""Gets the 4D closed orbit for a given dp

    Finds the closed orbit in the 4-d transverse phase space by numerically
    solving for a fixed point of the one turn map M calculated with
    :py:func:`.internal_lpass`.

    .. math:: \begin{pmatrix}x \\ p_x \\ y \\ p_y \\ dp \\ c\tau_2\end{pmatrix}
       =\mathbf{M} \cdot
       \begin{pmatrix}x \\ p_x \\ y \\ p_y \\ dp \\ c\tau_1\end{pmatrix}

    under the **CONSTANT MOMENTUM** constraint **dp** and with **NO**
    constraint on the 6-th coordinate :math:`c\tau`

    Important:
        :py:func:`find_orbit4` imposes a constraint on *dp* and relaxes the
        constraint on the revolution frequency. A physical storage ring does
        exactly the opposite: the momentum deviation of a particle on the
        closed orbit settles at the value such that the revolution is
        synchronous with the RF cavity: :math:`f_{RF} = h*f_{rev}`

    To impose this artificial constraint in :py:func:`find_orbit4`,
    the ``PassMethod`` used for any element **SHOULD NOT**:

    1. Change the longitudinal momentum *dp* (cavities ,
       magnets with radiation)
    2. Have any time dependence (localized impedance, fast kickers etc)

    Parameters:
        ring:           Lattice description (:py:attr:`~.Lattice.is_6d` must be
          :py:obj:`False`)
        dp:             Momentum deviation. Defaults to 0
        refpts:         Observation points.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        dct:            Path lengthening. If specified, *dp* is ignored and
          the off-momentum is deduced from the path lengthening.
        df:             Deviation from the nominal RF frequency. If specified,
          *dp* is ignored and the off-momentum is deduced from the frequency
          deviation.
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array). :py:func:`find_orbit4` propagates it to
          the specified *refpts*.
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: False

    Keyword Args:
        guess (Orbit):          (6,) initial value for the
          closed orbit. It may help convergence. Default: (0, 0, 0, 0, 0, 0)
        convergence (float):    Convergence criterion.
          Default: :py:data:`DConstant.OrbConvergence <.DConstant>`
        max_iterations (int):   Maximum number of iterations.
          Default: :py:data:`DConstant.OrbMaxIter <.DConstant>`
        XYStep (float):          Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`

    Returns:
        orbit0:         (6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py,dp,0)
        orbit:          (Nrefs, 6) closed orbit vector at each location
                        specified in *refpts*

    See also:
        :py:func:`find_sync_orbit`, :py:func:`find_orbit6`
    """
    if len([v for v in (dp, dct, df) if v is not None]) > 1:
        raise AtError("For off-momentum specification, only one of "
                      "dp, dct and df may be specified")
    if orbit is None:
        if df is not None:
            frf = ring.cell_revolution_frequency * ring.cell_harmnumber
            dct = -ring.cell_length * df / (frf+df)
            orbit = _orbit_dct(ring, dct, keep_lattice=keep_lattice, **kwargs)
        elif dct is not None:
            orbit = _orbit_dct(ring, dct, keep_lattice=keep_lattice, **kwargs)
        else:
            orbit = _orbit_dp(ring, dp, keep_lattice=keep_lattice, **kwargs)
        keep_lattice = True

    # bug in numpy < 1.13
    if ring.refcount(refpts) == 0:
        all_points = numpy.empty((0, 6), dtype=float)
    else:
        all_points = internal_lpass(ring, orbit.copy(order='K'), refpts=refpts,
                                    keep_lattice=keep_lattice)
        all_points = numpy.squeeze(all_points, axis=(1, 3)).T
    return orbit, all_points


def find_sync_orbit(ring: Lattice, dct: float = None, refpts: Refpts = None, *,
                    dp: float = None,
                    df: float = None,
                    orbit: Orbit = None,
                    keep_lattice: bool = False, **kwargs):
    r"""Gets the 4D closed orbit for a given dct

    Finds the closed orbit, synchronous with the RF cavity (first 5
    components of the phase space vector) by numerically solving for a fixed
    point of the one turn map M calculated with :py:func:`.internal_lpass`

    .. math:: \begin{pmatrix}x \\ p_x \\ y \\ p_y \\ dp \\ c\tau_1+
       dc\tau\end{pmatrix} =\mathbf{M} \cdot
       \begin{pmatrix}x \\ p_x \\ y \\ p_y \\ dp \\ c\tau_1\end{pmatrix}

    under the constraint :math:`dc\tau = c\tau_2 - c\tau_1 = c/f_{rev} -
    c/f_{rev_0}`, where :math:`f_{rev_0} = f_{RF_0}/h` is the design revolution
    frequency and :math:`f_{rev} = (f_{RF_0} + df_{RF})/h` is the imposed
    revolution frequency.

    Important:
        :py:func:`find_sync_orbit` imposes a constraint :math:`c\tau_2 -
        c\tau_1` and :math:`dp_2 = dp_1` but no constraint on the value of
        :math:`dp_2` or :math:`dp_1`. The algorithm assumes time-independent
        fixed-momentum ring to reduce the dimensionality of the problem.

    To impose this artificial constraint in :py:func:`find_sync_orbit`,
    the ``PassMethod`` used for any element **SHOULD NOT**:

    1. Change the longitudinal momentum *dp* (cavities ,
       magnets with radiation)
    2. Have any time dependence (localized impedance, fast kickers etc)

    Parameters:
        ring:           Lattice description (:py:attr:`~.Lattice.is_6d` must be
          :py:obj:`False`)
        dct:            Path lengthening.
        refpts:         Observation points.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        dp:             Momentum deviation. Defaults to :py:obj:`None`
        df:             Deviation from the nominal RF frequency. If specified,
          *dct* is ignored and the off-momentum is deduced from the frequency
          deviation.
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array). :py:func:`find_sync_orbit` propagates it
          to the specified *refpts*.
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: False

    Keyword Args:
        guess (Orbit):          (6,) initial value for the
          closed orbit. It may help convergence. Default: (0, 0, 0, 0, 0, 0)
        convergence (float):    Convergence criterion.
          Default: :py:data:`DConstant.OrbConvergence <.DConstant>`
        max_iterations (int):   Maximum number of iterations.
          Default: :py:data:`DConstant.OrbMaxIter <.DConstant>`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`

    Returns:
        orbit0:         (6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py,dp,0)
        orbit:          (Nrefs, 6) closed orbit vector at each location
                        specified in *refpts*

    See also:
        :py:func:`find_orbit4`, :py:func:`find_orbit6`
    """
    if len([v for v in (dp, dct, df) if v is not None]) > 1:
        raise AtError("For off-momentum specification, only one of "
                      "dp, dct and df may be specified")
    if orbit is None:
        if df is not None:
            frf = ring.cell_revolution_frequency * ring.cell_harmnumber
            dct = -ring.cell_length * df / (frf+df)
            orbit = _orbit_dct(ring, dct, keep_lattice=keep_lattice, **kwargs)
        elif dp is not None:
            orbit = _orbit_dp(ring, dp, keep_lattice=keep_lattice, **kwargs)
        else:
            orbit = _orbit_dct(ring, dct,  keep_lattice=keep_lattice, **kwargs)
        keep_lattice = True

    # bug in numpy < 1.13
    if ring.refcount(refpts) == 0:
        all_points = numpy.empty((0, 6), dtype=float)
    else:
        all_points = internal_lpass(ring, orbit.copy(order='K'), refpts=refpts,
                                    keep_lattice=keep_lattice)
        all_points = numpy.squeeze(all_points, axis=(1, 3)).T
    return orbit, all_points


def _orbit6(ring: Lattice, cavpts=None, guess=None, keep_lattice=False,
            **kwargs):
    """Solver for 6D motion"""
    convergence = kwargs.pop('convergence', DConstant.OrbConvergence)
    max_iterations = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    method = kwargs.pop('method', ELossMethod.TRACKING)
    rem = kwargs.keys()
    if len(rem) > 0:
        raise AtError(f'Unexpected keywords for orbit6: {", ".join(rem)}')

    l0 = get_s_pos(ring, len(ring))[0]
    f_rf = ring.get_rf_frequency()
    harm_number = round(f_rf*l0/ring.beta/clight)

    if guess is None:
        _, dt = get_timelag_fromU0(ring, method=method, cavpts=cavpts)
        # Getting timelag by tracking uses a different lattice,
        # so we cannot now use the same one again.
        if method is ELossMethod.TRACKING:
            keep_lattice = False
        ref_in = numpy.zeros((6,), order='F')
        ref_in[5] = -dt
    else:
        ref_in = numpy.copy(guess)

    theta = numpy.zeros((6,))
    theta[5] = ring.beta * clight * harm_number / f_rf - l0

    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0]) + \
        dp_step * numpy.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0])
    delta_matrix = numpy.asfortranarray(
        numpy.concatenate((numpy.diag(scaling), numpy.zeros((6, 1))), axis=1))

    id6 = numpy.asfortranarray(numpy.identity(6))
    change = 1
    itercount = 0
    while (change > convergence) and itercount < max_iterations:
        in_mat = ref_in.reshape((6, 1)) + delta_matrix
        _ = internal_lpass(ring, in_mat, refpts=[], keep_lattice=keep_lattice)
        # the reference particle after one turn
        ref_out = in_mat[:, 6]
        # 6x6 jacobian matrix from numerical differentiation:
        # f(x+d) - f(x) / d
        j6 = (in_mat[:, :6] - in_mat[:, 6:]) / scaling
        a = j6 - id6  # f'(r_n) - 1
        b = ref_out[:] - ref_in[:] - theta
        # b_over_a, _, _, _ = numpy.linalg.lstsq(a, b, rcond=-1)
        b_over_a = numpy.linalg.solve(a, b)
        r_next = ref_in - b_over_a
        # determine if we are close enough
        change = numpy.linalg.norm(r_next - ref_in)
        itercount += 1
        ref_in = r_next
        keep_lattice = True

    if itercount == max_iterations:
        warnings.warn(AtWarning('Maximum number of iterations reached. '
                                'Possible non-convergence'))
    return ref_in


# noinspection PyIncorrectDocstring
@frequency_control
def find_orbit6(ring: Lattice, refpts: Refpts = None, *,
                orbit: Orbit = None, keep_lattice: bool = False, **kwargs):
    r"""Gets the closed orbit in the full 6-D phase space

    Finds the closed orbit in the full 6-D phase space
    by numerically solving  for a fixed point of the one turn
    map M calculated with :py:func:`.internal_lpass`

    .. math:: \begin{pmatrix}x \\ p_x \\ y \\ p_y \\ dp \\ c\tau_2\end{pmatrix}
       =\mathbf{M} \cdot
       \begin{pmatrix}x \\ p_x \\ y \\ p_y \\ dp \\ c\tau_1\end{pmatrix}

    with constraint :math:`c\tau_2 - c\tau_1 = c.h (1/f_{RF} - 1/f_{RF_0})`

    Important:
        :py:func:`find_orbit6` is a realistic simulation:

        1.  The requency in the RF cavities :math:`f_{RF}` (may be different
            from :math:`f_{RF_0}`) imposes the synchronous condition
            :math:`c\tau_2 - c\tau_1 = c.h (1/f_{RF} - 1/f_{RF_0})`,
        2.  The algorithm numerically calculates the
            6-by-6 Jacobian matrix J6. In order for (J-E) matrix
            to be non-singular it is **NECESSARY** to use a realistic
            ``PassMethod`` for cavities with non-zero momentum kick
            (such as ``RFCavityPass``).
        3.  :py:func:`find_orbit6` can find orbits with radiation.
            In order for the solution to exist the cavity must supply an
            adequate energy compensation.
            In the simplest case of a single cavity, it must have
            ``Voltage`` set so that :math:`V > E_{loss}`, the energy loss
            per turn
        4.  There is a family of solutions that correspond to different RF
            buckets They differ in the 6-th coordinate by
            :math:`c*Nb/f_{RF}`,  Nb = 0:h-1
        5.  The value of the 6-th coordinate found at the cavity gives
            the equilibrium RF phase. If there is no radiation it is 0.
        6.  ``dp``, ``dct`` and ``df`` arguments are applied with respect
            to the **NOMINAL** on-momentum frequency. They overwrite
            exisiting frequency offsets
            

    Parameters:
        ring:           Lattice description
        refpts:         Observation points
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array). :py:func:`find_sync_orbit` propagates it
          to the specified *refpts*.
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: False

    Keyword Args:
        guess (Orbit):        (6,) initial value for the
          closed orbit. It may help convergence. Default: (0, 0, 0, 0, 0, 0)
        convergence (float):  Convergence criterion.
          Default: :py:data:`DConstant.OrbConvergence <.DConstant>`
        max_iterations (int): Maximum number of iterations.
          Default: :py:data:`DConstant.OrbMaxIter <.DConstant>`
        XYStep (float):       Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):       Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        method (ELossMethod): Method for energy loss computation.
          See :py:class:`.ELossMethod`.
        cavpts (Refpts):      Cavity location. If :py:obj:`None`, use all
          cavities. This is used to compute the initial synchronous phase.

    Returns:
        orbit0:         (6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py,dp,0)
        orbit:          (Nrefs, 6) closed orbit vector at each location
                        specified in *refpts*


    See also:
        :py:func:`find_orbit4`, :py:func:`find_sync_orbit`
    """
    if orbit is None:
        orbit = _orbit6(ring, keep_lattice=keep_lattice, **kwargs)
        keep_lattice = True

    # bug in numpy < 1.13
    if ring.refcount(refpts) == 0:
        all_points = numpy.empty((0, 6), dtype=float)
    else:
        all_points = internal_lpass(ring, orbit.copy(order='K'), refpts=refpts,
                                    keep_lattice=keep_lattice)
        all_points = numpy.squeeze(all_points, axis=(1, 3)).T
    return orbit, all_points


def find_orbit(ring, refpts: Refpts = None, **kwargs):
    r"""Gets the closed orbit in the general case

    Depending on the lattice, :py:func:`find_orbit` will:

    * use :py:func:`find_orbit6` if :py:attr:`~.Lattice.is_6d` is
      :py:obj:`True`,
    * use :py:func:`find_sync_orbit` if :py:attr:`~.Lattice.is_6d` is
      :py:obj:`False` and
      *dct* or *df* is specified,
    * use :py:func:`find_orbit4` otherwise.

    Parameters:
        ring:           Lattice description
        refpts:         Observation points.
          See ":ref:`Selecting elements in a lattice <refpts>`"

    Keyword Args:
        orbit (Orbit):          Avoids looking for initial the closed
          orbit if it is already known. :py:func:`find_orbit` propagates it
          to the specified *refpts*.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation from the nominal RF frequency.
          Defaults to :py:obj:`None`
        guess (Orbit):          (6,) initial value for the closed orbit.
          It may help convergence. Default: (0, 0, 0, 0, 0, 0)
        convergence (float):    Convergence criterion.
          Default: :py:data:`DConstant.OrbConvergence <.DConstant>`
        max_iterations (int):   Maximum number of iterations.
          Default: :py:data:`DConstant.OrbMaxIter <.DConstant>`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    For other keywords, refer to the underlying methods

    Returns:
        orbit0:         (6,) closed orbit vector at the entrance of the
                        1-st element (x,px,y,py,dp,0)
        orbit:          (Nrefs, 6) closed orbit vector at each location
                        specified in *refpts*

    See also:
        :py:func:`find_orbit4`, :py:func:`find_sync_orbit`,
        :py:func:`find_orbit6`
    """
    if ring.is_6d:
        return find_orbit6(ring, refpts=refpts, **kwargs)
    else:
        return find_orbit4(ring, refpts=refpts, **kwargs)


Lattice.find_orbit4 = find_orbit4
Lattice.find_sync_orbit = find_sync_orbit
Lattice.find_orbit6 = find_orbit6
Lattice.find_orbit = find_orbit



"""
Tracking functions
"""
from .atpass import reset_rng, common_rng, thread_rng
from .track import *
from .particles import *
from .utils import *
from .deprecated import *
# initialise the C random generators
reset_rng()



import numpy
import pytest
from numpy.testing import assert_allclose, assert_equal

from at import elements
from at.lattice import Lattice, AtWarning, AtError


def test_lattice_creation_gets_attributes_from_arguments():
    lat = Lattice(name="lattice", energy=3.0e6, periodicity=32, an_attr=12)
    assert len(lat) == 0
    assert lat.name == "lattice"
    assert lat.energy == 3.0e6
    assert lat.periodicity == 32
    assert lat.radiation is False
    assert lat.particle.name == "relativistic"
    assert lat.an_attr == 12


def test_lattice_energy_radiation_periodicity():
    d = elements.Dipole(
        "d1",
        1,
        BendingAngle=numpy.pi / 16,
        Energy=5.0e6,
        PassMethod="BndMPoleSymplectic4RadPass",
    )
    lat = Lattice([d], name="lattice", energy=3.0e6)
    assert lat.energy == 3.0e6
    assert lat.periodicity == 32
    assert lat.radiation is True
    assert lat.is_6d is True


def test_lattice_voltage_harmonic_number():
    rf = elements.RFCavity("rf", 0, 0.2e6, 0.5e9, 5, 3.0e6)
    d = elements.Dipole("d1", 2.99792458, BendingAngle=numpy.pi / 5)
    lat = Lattice([rf, d], name="lattice")
    assert lat.energy == 3.0e6
    assert lat.periodicity == 10
    assert lat.rf_voltage == 2e6
    assert lat.revolution_frequency == 10.0e6
    assert lat.harmonic_number == 50
    assert lat.radiation is True


def test_lattice_creation_from_lattice_inherits_attributes():
    d = elements.Dipole(
        "d1",
        1,
        BendingAngle=numpy.pi,
        Energy=5.0e6,
        PassMethod="BndMPoleSymplectic4RadPass",
    )
    lat1 = Lattice([d], name="lattice", energy=3.0e6, periodicity=32, an_attr=12)
    lat2 = Lattice(lat1, another_attr=5)
    assert id(lat1) != id(lat2)
    assert len(lat2) == 1
    assert lat2.name == "lattice"
    assert lat2.energy == 3.0e6
    assert lat2.periodicity == 32
    assert lat2.is_6d is True
    assert lat2.another_attr == 5
    with pytest.raises(AttributeError):
        assert lat2.an_attr == 12


def test_lattice_energy_is_not_defined_raises_AtError():
    d = elements.Dipole("d1", 1, BendingAngle=numpy.pi)
    with pytest.raises(AtError):
        Lattice([d])


def test_item_is_not_an_AT_element_warns_correctly():
    with pytest.warns(AtWarning):
        Lattice(["a"], energy=0, periodicity=1)


def test_lattice_string_ordering():
    lat = Lattice(
        [elements.Drift("D0", 1.0, attr1=numpy.array(0))],
        name="lat",
        energy=5,
        periodicity=1,
        attr2=3,
    )
    latstr = str(lat)
    assert latstr.startswith(
        "Lattice(<1 elements>, name='lat', "
        "energy=5, particle=Particle('relativistic'), "
        "periodicity=1, beam_current=0.0, nbunch=1"
    )
    assert latstr.endswith("attr2=3)")

    latrepr = repr(lat)
    assert latrepr.startswith(
        "Lattice([Drift('D0', 1.0, attr1=array(0))], "
        "name='lat', "
        "energy=5, particle=Particle('relativistic'), "
        "periodicity=1, beam_current=0.0, nbunch=1"
    )
    assert latrepr.endswith("attr2=3)")


def test_getitem(simple_lattice, simple_ring):
    assert simple_lattice[-1] == simple_ring[-1]
    bool_refs = numpy.array([False, True, False, True, False, True])
    assert simple_lattice[bool_refs] == simple_ring[1::2]


def test_setitem(simple_lattice):
    new = elements.Monitor("M2")
    old = simple_lattice[5]
    simple_lattice[5] = new
    assert simple_lattice[5] != old
    assert simple_lattice[5] == new
    bool_refs = numpy.array([False, False, False, False, False, True])
    simple_lattice[bool_refs] = old
    assert simple_lattice[5] != new
    assert simple_lattice[5] == old


def test_delitem(simple_lattice, simple_ring):
    mon = elements.Monitor("M2")
    simple_lattice.append(mon)
    assert len(simple_lattice) == 7
    del simple_lattice[-1]
    assert simple_lattice[:] == simple_ring
    bool_refs = numpy.array([False, False, False, False, False, False, True])
    simple_lattice.append(mon)
    assert len(simple_lattice) == 7
    del simple_lattice[bool_refs]
    assert simple_lattice[:] == simple_ring


def test_copy(hmba_lattice):
    assert id(hmba_lattice.copy()) != id(hmba_lattice)
    assert id(hmba_lattice.copy()[0]) == id(hmba_lattice[0])


def test_deepcopy(hmba_lattice):
    assert id(hmba_lattice.deepcopy()) != id(hmba_lattice)
    assert id(hmba_lattice.deepcopy()[0]) != id(hmba_lattice[0])


def test_property_values_against_known(hmba_lattice):
    assert hmba_lattice.rf_voltage == 6000000
    assert hmba_lattice.harmonic_number == 992
    assert hmba_lattice.radiation is False
    numpy.testing.assert_almost_equal(
        hmba_lattice.energy_loss, 2526188.658758993, decimal=1
    )


def test_radiation_change(hmba_lattice):
    rfs = [elem for elem in hmba_lattice if isinstance(elem, elements.RFCavity)]
    dipoles = [elem for elem in hmba_lattice if isinstance(elem, elements.Dipole)]
    quads = [elem for elem in hmba_lattice if isinstance(elem, elements.Quadrupole)]
    hmba_lattice.radiation_on(None, "pass2", "auto")
    assert hmba_lattice.radiation is True
    assert hmba_lattice.has_cavity is False
    for elem in rfs:
        assert elem.PassMethod == "IdentityPass"
    for elem in dipoles:
        assert elem.PassMethod == "pass2"
    for elem in quads:
        assert elem.PassMethod == "StrMPoleSymplectic4RadPass"
    hmba_lattice.radiation_off(None, "BndMPoleSymplectic4Pass", "auto")
    assert hmba_lattice.radiation is False
    for elem in rfs:
        assert elem.PassMethod == "IdentityPass"
    for elem in dipoles:
        assert elem.PassMethod == "BndMPoleSymplectic4Pass"
    for elem in quads:
        assert elem.PassMethod == "StrMPoleSymplectic4Pass"


def test_radiation_state_errors(hmba_lattice):
    hmba_lattice.radiation_on()
    with pytest.raises(AtError):
        hmba_lattice.linopt()
    hmba_lattice.radiation_off()
    hmba_lattice.linopt()
    with pytest.raises(AtError):
        hmba_lattice.ohmi_envelope()
    hmba_lattice.enable_6d()
    hmba_lattice.ohmi_envelope()
    with pytest.raises(AtError):
        hmba_lattice.get_mcf()
    hmba_lattice.disable_6d()
    hmba_lattice.get_mcf()


@pytest.mark.parametrize("ring", ["hmba_lattice"])
def test_develop(request, ring):
    ring = request.getfixturevalue(ring)
    newring = ring.develop()
    assert_allclose(ring.circumference, newring.circumference)
    assert_allclose(ring.rf_voltage, newring.rf_voltage)
    assert_allclose(
        ring.revolution_frequency * ring.harmonic_number,
        newring.revolution_frequency * newring.harmonic_number,
    )
    rp1 = ring.radiation_parameters()
    rp2 = newring.radiation_parameters()
    assert_allclose(rp1.fulltunes, rp2.fulltunes)
    assert_allclose(rp1.U0, rp2.U0)


@pytest.mark.parametrize("ring", ["hmba_lattice"])
def test_operators(request, ring):
    ring = request.getfixturevalue(ring)
    newring1 = ring + ring
    newring2 = ring.concatenate(ring, copy=True)
    assert_equal(len(newring1), len(ring) * 2)
    assert_equal(len(newring2), len(ring) * 2)
    newring1 += ring
    newring2.concatenate(ring, copy=False)
    assert_equal(len(newring1), len(ring) * 3)
    assert_equal(len(newring2), len(ring) * 3)
    newring1 = ring * 2
    newring2 = ring.repeat(2)
    assert_equal(len(newring1), len(ring) * 2)
    assert_equal(len(newring2), len(ring) * 2)
    assert_equal(newring1.harmonic_number, ring.harmonic_number)
    newring1 = ring.reverse(copy=True)
    assert_equal(newring1[-1].FamName, ring[0].FamName)



import at
import numpy
import pytest
import warnings
from at import AtWarning
from at.lattice import Lattice, elements, params_filter, no_filter
from at.load.utils import find_class, element_from_dict
# noinspection PyProtectedMember
from at.load.utils import _CLASS_MAP, _PASS_MAP
from at.load.utils import RingParam, split_ignoring_parentheses
# noinspection PyProtectedMember
from at.load.matfile import ringparam_filter


def _matlab_scanner(element_list, **kwargs):
    """This function simulates a mat-file reading but replaces the mat-file by
    a list of elements"""
    latt = Lattice(ringparam_filter, no_filter, element_list,
                   iterator=params_filter, **kwargs)
    return vars(latt)


def test_lattice_gets_attributes_from_RingParam():
    p = RingParam('lattice_name', 3.e+6, Periodicity=32)
    params = _matlab_scanner([p])
    assert params['name'] == 'lattice_name'
    assert params['_energy'] == 3.e+6
    assert params['periodicity'] == 32
    assert params['_radiation'] is False


def test_lattice_gets_attributes_from_elements():
    d = elements.Dipole('d1', 1, BendingAngle=numpy.pi, Energy=3.e+6,
                        PassMethod='BndMPoleSymplectic4RadPass')
    params = _matlab_scanner([d])
    assert params['name'] == ''
    assert params['_energy'] == 3.e+6
    assert params['periodicity'] == 2
    assert params['_radiation'] is True


def test_no_bending_in_the_cell_warns_correctly():
    with pytest.warns(AtWarning):
        params = _matlab_scanner([], energy=3.e+9)
        assert params['periodicity'] == 1


def test_non_integer_number_of_cells_warns_correctly():
    d = elements.Dipole('d1', 1, BendingAngle=0.195)
    with pytest.warns(AtWarning):
        params = _matlab_scanner([d], energy=3.e+9)
        assert params['periodicity'] == 32


def test_inconsistent_energy_values_warns_correctly():
    d1 = elements.Dipole('d1', 1, BendingAngle=numpy.pi, Energy=5.e+9,
                         PassMethod='BndMPoleSymplectic4RadPass')
    d2 = elements.Dipole('d2', 1, BendingAngle=numpy.pi, Energy=3.e+9,
                         PassMethod='BndMPoleSymplectic4RadPass')
    m1 = elements.Marker('m1', Energy=5.e+9)
    m2 = elements.Marker('m2', Energy=3.e+9)
    with pytest.warns(AtWarning):
        params = _matlab_scanner([m1, m2])
        assert params['_energy'] == 5.e+9
    with pytest.warns(AtWarning):
        params = _matlab_scanner([d1, d2])
        assert params['_energy'] == 5.e+9


def test_more_than_one_RingParam_in_ring_raises_warning():
    p1 = RingParam('rp1', 5.e+6, 1)
    p2 = RingParam('rp2', 3.e+6, 1)
    with pytest.warns(AtWarning):
        params = _matlab_scanner([p1, p2])
        assert params['_energy'] == 5.e+6


def test_invalid_class_warns_correctly():
    elem_kwargs = {'Class': 'Invalid'}
    with pytest.warns(at.AtWarning):
        find_class(elem_kwargs, quiet=False)
    with warnings.catch_warnings(record=True) as record:
        find_class(elem_kwargs, quiet=True)
    assert len(record) == 0


def test_no_pass_method_warns_correctly():
    elem_kwargs = {}
    with pytest.warns(at.AtWarning):
        find_class(elem_kwargs, quiet=False)
    with warnings.catch_warnings(record=True) as record:
        find_class(elem_kwargs, quiet=True)
    assert len(record) == 0


def test_invalid_pass_method_warns_correctly():
    elem_kwargs = {'PassMethod': 'invalid'}
    with pytest.warns(at.AtWarning):
        find_class(elem_kwargs, quiet=False)
    with warnings.catch_warnings(record=True) as record:
        find_class(elem_kwargs, quiet=True)
    assert len(record) == 0


def test_class_mapping():
    elem_kwargs = {'FamName': 'fam'}
    for class_name in _CLASS_MAP:
        elem_kwargs['Class'] = class_name
        assert find_class(elem_kwargs) is _CLASS_MAP[class_name]
        elem_kwargs['Class'] = class_name.upper()
        assert find_class(elem_kwargs) is _CLASS_MAP[class_name]


def test_family_mapping():
    elem_kwargs = {}
    for family_name in _CLASS_MAP:
        elem_kwargs['FamName'] = family_name
        assert find_class(elem_kwargs) is _CLASS_MAP[family_name]
        elem_kwargs['FamName'] = family_name.upper()
        assert find_class(elem_kwargs) is _CLASS_MAP[family_name]


def test_PassMethod_mapping():
    elem_kwargs = {'FamName': 'fam'}
    for pass_method in _PASS_MAP.keys():
        elem_kwargs['PassMethod'] = pass_method
        assert find_class(elem_kwargs) == _PASS_MAP[pass_method]


def test_find_Aperture():
    elem_kwargs = {'Limits': [-0.5, 0.5, -0.5, 0.5], 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.Aperture


@pytest.mark.parametrize('elem_kwargs', (
        {'Voltage': 2.5e+6, 'FamName': 'fam'},
        {'Frequency': 5.e+8, 'FamName': 'fam'},
        {'HarmNumber': 1000, 'FamName': 'fam'},
        {'PhaseLag': 0, 'FamName': 'fam'},
        {'TimeLag': 0.0, 'FamName': 'fam'}))
def test_find_RFCavity(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.RFCavity


def test_find_Monitor():
    elem_kwargs = {'GCR': [1, 1, 0, 0], 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.Monitor


@pytest.mark.parametrize('elem_kwargs', (
        {'FullGap': 0.05, 'FamName': 'fam'},
        {'FringeInt1': 0.5, 'FamName': 'fam'},
        {'FringeInt2': 0.5, 'FamName': 'fam'},
        {'gK': 0.03, 'FamName': 'fam'},
        {'EntranceAngle': 0.05, 'FamName': 'fam'},
        {'ExitAngle': 0.05, 'FamName': 'fam'}))
def test_find_Dipole(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.Dipole


def test_find_Corrector():
    elem_kwargs = {'KickAngle': [0, 0], 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.Corrector


def test_find_RingParam():
    elem_kwargs = {'Periodicity': 1, 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is RingParam


def test_find_M66():
    elem_kwargs = {'M66': numpy.eye(6), 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.M66


@pytest.mark.parametrize('elem_kwargs', (
        {'K': -0.5, 'FamName': 'fam'},
        {'PolynomB': [0, 1, 0, 0], 'FamName': 'fam'}))
def test_find_Quadrupole(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.Quadrupole


@pytest.mark.parametrize('elem_kwargs', (
        {'PolynomB': [0, 0, 0, 0], 'PassMethod': 'StrMPoleSymplectic4Pass',
         'FamName': 'fam'},
        {'PolynomB': [0, 0, 0, 0], 'Length': 1, 'FamName': 'fam'}))
def test_find_Multipole(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.Multipole


def test_find_Drift():
    elem_kwargs = {'Length': 1, 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.Drift


def test_find_Sextupole():
    elem_kwargs = {'PolynomB': [0, 0, 1, 0], 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.Sextupole


def test_find_Octupole():
    elem_kwargs = {'PolynomB': [0, 0, 0, 1], 'FamName': 'fam'}
    assert find_class(elem_kwargs, True) is elements.Octupole


@pytest.mark.parametrize('elem_kwargs', (
        {'PolynomB': [0, 0, 0, 0], 'FamName': 'fam'},
        {'PolynomB': [0, 0, 0, 0, 1], 'FamName': 'fam'},
        {'PolynomB': [0, 0, 0, 0], 'FamName': 'fam'},
        {'PolynomB': [0, 0, 0, 0, 1], 'Length': 0, 'FamName': 'fam'}))
def test_find_ThinMultipole(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.ThinMultipole


@pytest.mark.parametrize('elem_kwargs', (
        {'FamName': 'fam', 'PassMethod': 'IdentityPass'},
        {'Length': 0.0, 'FamName': 'fam', 'PassMethod': 'IdentityPass'}))
def test_find_Marker(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.Marker


@pytest.mark.parametrize('elem_kwargs', (
        {'FamName': 'fam'},
        {'Length': 0.0, 'FamName': 'fam'}))
def test_find_Element(elem_kwargs):
    assert find_class(elem_kwargs, True) is elements.Element


@pytest.mark.parametrize('elem_kwargs', (
        {'Class': 'Marker', 'PassMethod': 'IdentityPass', 'Length': 1,
         'FamName': 'fam'},
        {'Class': 'Quadrupole', 'PassMethod': 'CavityPass', 'Length': 1,
         'FamName': 'fam'},
        {'Class': 'Marker', 'PassMethod': 'StrMPoleSymplectic4Pass',
         'FamName': 'fam'},
        {'Class': 'Drift', 'PassMethod': 'StrMPoleSymplectic4Pass',
         'Length': 1.0, 'FamName': 'fam'},
        {'Class': 'Drift', 'PassMethod': 'InvalidPass',
         'Length': 1.0, 'FamName': 'fam'}))
def test_sanitise_class_error(elem_kwargs):
    with pytest.warns(AtWarning):
        element_from_dict(elem_kwargs)


@pytest.mark.parametrize(
    "string,delimiter,target",
    [
        ["a,b", ",", ["a", "b"]],
        ["a,b(c,d)", ",", ["a", "b(c,d)"]],
        ["l=0,hom(4,0.0,0)", ",", ["l=0", "hom(4,0.0,0)"]],
        ["inv(arca_c1r),3*(ms,arca_c2)", ",",
         ["inv(arca_c1r)", "3*(ms,arca_c2)"]]
    ],
)
def test_split_ignoring_parentheses(string, delimiter, target):
    assert split_ignoring_parentheses(string, delimiter) == target
